import copy
import inspect
from typing import Any, Callable, Dict, Iterable, Optional

from triad import extension_method
from triad.collections.function_wrapper import (
    AnnotatedParam,
    FunctionWrapper,
    function_wrapper,
)
from triad.utils.assertion import assert_or_throw
from triad.utils.convert import get_caller_global_local_vars, to_function

from fugue.constants import FUGUE_ENTRYPOINT
from fugue.exceptions import FugueInterfacelessError
from fugue.workflow.workflow import FugueWorkflow, WorkflowDataFrame, WorkflowDataFrames


def module(
    func: Optional[Callable] = None,
    as_method: bool = False,
    name: Optional[str] = None,
    on_dup: str = "overwrite",
) -> Any:
    """Decorator for module

    Please read :doc:`Module Tutorial <tutorial:tutorials/module>`
    """

    if func is None:  # @module(...)
        return lambda func: module(func, as_method=as_method, name=name, on_dup=on_dup)
    # @module
    res = _ModuleFunctionWrapper(func)
    if as_method:
        extension_method(func, name=name, on_dup=on_dup)
    return res


def _to_module(
    obj: Any,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> "_ModuleFunctionWrapper":
    if isinstance(obj, _ModuleFunctionWrapper):
        return obj
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, _ModuleFunctionWrapper):
            return copy.copy(f)
        # this is for functions without decorator
        return _ModuleFunctionWrapper(f)
    except Exception as e:
        exp = e
    raise FugueInterfacelessError(f"{obj} is not a valid module", exp)


@function_wrapper(FUGUE_ENTRYPOINT)
class _ModuleFunctionWrapper(FunctionWrapper):
    def __init__(  # pylint: disable-all
        self,
        func: Callable,
        params_re: str = "^(w?(u|v+)|w(u?|v*))x*z?$",
        return_re: str = "^[uvn]$",
    ):
        super().__init__(func, params_re, return_re)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self._need_add_workflow(*args, **kwargs):
            wf = self._infer_workflow(*args, **kwargs)
            assert_or_throw(wf is not None, ValueError("can't infer workflow"))
            return super().__call__(wf, *args, **kwargs)
        return super().__call__(*args, **kwargs)

    @property
    def has_input(self) -> bool:
        return any(
            isinstance(x, (_WorkflowDataFrameParam, _WorkflowDataFramesParam))
            for x in self._params.values()
        )

    @property
    def has_dfs_input(self) -> bool:
        return any(
            isinstance(x, _WorkflowDataFramesParam) for x in self._params.values()
        )

    @property
    def has_no_output(self) -> bool:
        return not isinstance(
            self._rt, (_WorkflowDataFrameParam, _WorkflowDataFramesParam)
        )

    @property
    def has_single_output(self) -> bool:
        return isinstance(self._rt, _WorkflowDataFrameParam)

    @property
    def has_multiple_output(self) -> bool:
        return isinstance(self._rt, _WorkflowDataFramesParam)

    @property
    def _first_annotation_is_workflow(self) -> bool:
        return isinstance(self._params.get_value_by_index(0), _FugueWorkflowParam)

    def _need_add_workflow(self, *args: Any, **kwargs: Any):
        if not self._first_annotation_is_workflow:
            return False
        if self._params.get_key_by_index(0) in kwargs:
            return False
        if len(args) > 0 and isinstance(args[0], FugueWorkflow):
            return False
        return True

    def _infer_workflow(self, *args: Any, **kwargs: Any) -> Optional[FugueWorkflow]:
        def select_args() -> Iterable[Any]:
            for a in args:
                if isinstance(a, (WorkflowDataFrames, WorkflowDataFrame)):
                    yield a
            for _, v in kwargs.items():
                if isinstance(v, (WorkflowDataFrames, WorkflowDataFrame)):
                    yield v

        wf: Optional[FugueWorkflow] = None
        for a in select_args():
            if isinstance(a, WorkflowDataFrame):
                assert_or_throw(
                    wf is None or a.workflow is wf,
                    ValueError("different parenet workflows found on input dataframes"),
                )
                wf = a.workflow
            elif isinstance(a, WorkflowDataFrames):
                for k, v in a.items():
                    assert_or_throw(
                        isinstance(v, WorkflowDataFrame),
                        lambda: ValueError(f"{k}:{v} is not a WorkflowDataFrame"),
                    )
                    assert_or_throw(
                        wf is None or v.workflow is wf,
                        ValueError(
                            "different parenet workflows found on input dataframes"
                        ),
                    )
                    wf = v.workflow
        return wf


@_ModuleFunctionWrapper.annotated_param(
    FugueWorkflow,
    "w",
    matcher=lambda x: inspect.isclass(x) and issubclass(x, FugueWorkflow),
)
class _FugueWorkflowParam(AnnotatedParam):
    pass


@_ModuleFunctionWrapper.annotated_param(WorkflowDataFrame, "v")
class _WorkflowDataFrameParam(AnnotatedParam):
    pass


@_ModuleFunctionWrapper.annotated_param(WorkflowDataFrames, "u")
class _WorkflowDataFramesParam(AnnotatedParam):
    pass



from pytest import raises

import fugue.api as fa
from fugue import NativeExecutionEngine, register_global_conf
from fugue.exceptions import FugueInvalidOperation


class MyEngine(NativeExecutionEngine):
    def __init__(self, conf=None):
        super().__init__(conf)
        self.pre_enter_state = []
        self.post_exit_state = []
        self.stop_calls = 0

    def on_enter_context(self) -> None:
        self.pre_enter_state += [self.in_context]

    def on_exit_context(self) -> None:
        self.post_exit_state += [self.in_context]

    def stop_engine(self) -> None:
        self.stop_calls += 1


def test_engine_operations():
    o = MyEngine()
    assert fa.get_current_conf().get("fugue.x", 0) == 0
    register_global_conf({"fugue.x": 1})
    assert fa.get_current_conf().get("fugue.x", 0) == 1
    e = fa.set_global_engine(o, {"fugue.x": 2})
    assert e.pre_enter_state == [False]
    assert e.post_exit_state == []
    assert fa.get_current_conf().get("fugue.x", 0) == 2
    assert isinstance(e, NativeExecutionEngine)
    assert e.in_context and e.is_global
    assert fa.get_context_engine() is e
    with fa.engine_context("duckdb", {"fugue.x": 3}) as e2:
        assert fa.get_current_conf().get("fugue.x", 0) == 3
        assert fa.get_context_engine() is e2
        assert not e2.is_global and e2.in_context
        with e.as_context():
            assert e.pre_enter_state == [False, True]
            assert e.post_exit_state == []
            assert fa.get_current_conf().get("fugue.x", 0) == 2
            assert not e2.is_global and e2.in_context
            assert e.in_context and e.is_global
            assert fa.get_context_engine() is e
            assert e.stop_calls == 0
        assert e.pre_enter_state == [False, True]
        assert e.post_exit_state == [True]
        assert fa.get_current_conf().get("fugue.x", 0) == 3
        assert e.in_context and e.is_global
        assert fa.get_context_engine() is e2
    assert e.stop_calls == 0
    assert e.pre_enter_state == [False, True]
    assert e.post_exit_state == [True]
    assert fa.get_current_conf().get("fugue.x", 0) == 2
    assert not e2.is_global and not e2.in_context
    assert e.in_context and e.is_global
    e3 = fa.set_global_engine("duckdb", {"fugue.x": 4})
    assert e.stop_calls == 1
    assert e.pre_enter_state == [False, True]
    assert e.post_exit_state == [True, False]
    assert fa.get_current_conf().get("fugue.x", 0) == 4
    assert not e.in_context and not e.is_global
    assert e3.in_context and e3.is_global
    fa.clear_global_engine()
    assert not e3.in_context and not e3.is_global
    assert fa.get_current_conf().get("fugue.x", 0) == 1
    raises(FugueInvalidOperation, lambda: fa.get_context_engine())



from typing import Any, Dict, Iterable, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
import ray
import ray.data as rd
from triad.collections.schema import Schema

from fugue.dataframe import ArrowDataFrame, DataFrame, LocalBoundedDataFrame
from fugue.dataframe.dataframe import _input_schema
from fugue.exceptions import FugueDataFrameOperationError, FugueDatasetEmptyError
from fugue.plugins import (
    as_local_bounded,
    get_column_names,
    get_num_partitions,
    is_df,
    rename,
)

from ._constants import _ZERO_COPY
from ._utils.dataframe import build_empty, get_dataset_format


class RayDataFrame(DataFrame):
    """DataFrame that wraps Ray DataSet. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: :class:`ray:ray.data.Dataset`, :class:`pa:pyarrow.Table`,
      :class:`pd:pandas.DataFrame`,
      Fugue :class:`~fugue.dataframe.dataframe.DataFrame`,
      or list or iterable of arrays
    :param schema: |SchemaLikeObject|, defaults to None. If the schema
      is different from the ``df`` schema, then type casts will happen.
    :param internal_schema: for internal schema, it means the schema
      is guaranteed by the provider to be consistent with the schema of
      ``df``, so no type cast will happen. Defaults to False. This is
      for internal use only.
    """

    def __init__(  # noqa: C901
        self,
        df: Any = None,
        schema: Any = None,
        internal_schema: bool = False,
    ):
        metadata: Any = None
        if internal_schema:
            schema = _input_schema(schema).assert_not_empty()
        if df is None:
            schema = _input_schema(schema).assert_not_empty()
            super().__init__(schema)
            self._native = build_empty(schema)
            return
        if isinstance(df, rd.Dataset):
            fmt = get_dataset_format(df)
            if fmt is None:  # empty:
                schema = _input_schema(schema).assert_not_empty()
                super().__init__(schema)
                self._native = build_empty(schema)
                return
            elif fmt == "pandas":
                rdf = rd.from_arrow_refs(df.to_arrow_refs())
            elif fmt == "arrow":
                rdf = df
            else:
                raise NotImplementedError(
                    f"Ray Dataset in {fmt} format is not supported"
                )
        elif isinstance(df, pa.Table):
            rdf = rd.from_arrow(df)
            if schema is None:
                schema = df.schema
        elif isinstance(df, RayDataFrame):
            rdf = df._native
            if schema is None:
                schema = df.schema
            metadata = None if not df.has_metadata else df.metadata
        elif isinstance(df, (pd.DataFrame, pd.Series)):
            if isinstance(df, pd.Series):
                df = df.to_frame()
            adf = ArrowDataFrame(df)
            rdf = rd.from_arrow(adf.native)
            if schema is None:
                schema = adf.schema
        elif isinstance(df, Iterable):
            schema = _input_schema(schema).assert_not_empty()
            t = ArrowDataFrame(df, schema)
            rdf = rd.from_arrow(t.as_arrow())
        elif isinstance(df, DataFrame):
            rdf = rd.from_arrow(df.as_arrow(type_safe=True))
            if schema is None:
                schema = df.schema
            metadata = None if not df.has_metadata else df.metadata
        else:
            raise ValueError(f"{df} is incompatible with RayDataFrame")
        rdf, schema = self._apply_schema(rdf, schema, internal_schema)
        super().__init__(schema)
        self._native = rdf
        if metadata is not None:
            self.reset_metadata(metadata)

    @property
    def native(self) -> rd.Dataset:
        """The wrapped ray Dataset"""
        return self._native

    def native_as_df(self) -> rd.Dataset:
        return self._native

    @property
    def is_local(self) -> bool:
        return False

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        adf = self.as_arrow()
        if adf.shape[0] == 0:
            res = ArrowDataFrame([], self.schema)
        else:
            res = ArrowDataFrame(adf)
        if self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    @property
    def is_bounded(self) -> bool:
        return True

    @property
    def empty(self) -> bool:
        return len(self.native.take(1)) == 0

    @property
    def num_partitions(self) -> int:
        return _rd_num_partitions(self.native)

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        cols = (self.schema - cols).names
        return self._select_cols(cols)

    def _select_cols(self, cols: List[Any]) -> DataFrame:
        if cols == self.columns:
            return self
        rdf = self.native.map_batches(
            lambda b: b.select(cols),
            batch_format="pyarrow",
            **_ZERO_COPY,
            **self._remote_args(),
        )
        return RayDataFrame(rdf, self.schema.extract(cols), internal_schema=True)

    def peek_array(self) -> List[Any]:
        data = self.native.limit(1).to_pandas().values.tolist()
        if len(data) == 0:
            raise FugueDatasetEmptyError
        return data[0]

    def persist(self, **kwargs: Any) -> "RayDataFrame":
        # TODO: it mutates the dataframe, is this a good bahavior
        if not self.native.is_fully_executed():  # pragma: no cover
            self.native.fully_executed()
        return self

    def count(self) -> int:
        return self.native.count()

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return pa.concat_tables(_get_arrow_tables(self.native))

    def as_pandas(self) -> pd.DataFrame:
        return self.as_arrow().to_pandas()

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            new_schema = self.schema.rename(columns)
            new_cols = new_schema.names
        except Exception as e:
            raise FugueDataFrameOperationError from e
        rdf = self.native.map_batches(
            lambda b: b.rename_columns(new_cols),
            batch_format="pyarrow",
            **_ZERO_COPY,
            **self._remote_args(),
        )
        return RayDataFrame(rdf, schema=new_schema, internal_schema=True)

    def alter_columns(self, columns: Any) -> DataFrame:
        def _alter(
            table: pa.Table,
        ) -> pa.Table:  # pragma: no cover (pytest can't capture)
            return ArrowDataFrame(table).alter_columns(columns).native  # type: ignore

        new_schema = self._get_altered_schema(columns)
        if self.schema == new_schema:
            return self
        rdf = self.native.map_batches(
            _alter, batch_format="pyarrow", **_ZERO_COPY, **self._remote_args()
        )
        return RayDataFrame(rdf, schema=new_schema, internal_schema=True)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        df: DataFrame = self
        if columns is not None:
            df = df[columns]
        adf = df.as_arrow()
        if adf.shape[0] == 0:
            return []
        return ArrowDataFrame(adf).as_array(type_safe=type_safe)

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        yield from self.as_array(columns=columns, type_safe=type_safe)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        if columns is not None:
            return self[columns].head(n)
        pdf = RayDataFrame(self.native.limit(n), schema=self.schema)
        return pdf.as_local()  # type: ignore

    def _apply_schema(
        self, rdf: rd.Dataset, schema: Optional[Schema], internal_schema: bool
    ) -> Tuple[rd.Dataset, Schema]:
        if internal_schema:
            return rdf, schema
        if get_dataset_format(rdf) is None:  # empty
            schema = _input_schema(schema).assert_not_empty()
            return build_empty(schema), schema
        if schema is None or schema == rdf.schema(fetch_if_missing=True):
            return rdf, rdf.schema(fetch_if_missing=True)

        def _alter(table: pa.Table) -> pa.Table:  # pragma: no cover
            return ArrowDataFrame(table).alter_columns(schema).native  # type: ignore

        return (
            rdf.map_batches(
                _alter, batch_format="pyarrow", **_ZERO_COPY, **self._remote_args()
            ),
            schema,
        )

    def _remote_args(self) -> Dict[str, Any]:
        return {"num_cpus": 1}


@is_df.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_is_df(df: rd.Dataset) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_num_partitions(df: rd.Dataset) -> int:
    return df.num_blocks()


@as_local_bounded.candidate(lambda df: isinstance(df, rd.Dataset))
def _rd_as_local(df: rd.Dataset) -> bool:
    return pa.concat_tables(_get_arrow_tables(df))


@get_column_names.candidate(lambda df: isinstance(df, rd.Dataset))
def _get_ray_dataframe_columns(df: rd.Dataset) -> List[Any]:
    fmt = get_dataset_format(df)
    if fmt == "pandas":
        return list(df.schema(True).names)
    elif fmt == "arrow":
        return [f.name for f in df.schema(True)]
    raise NotImplementedError(f"{fmt} is not supported")  # pragma: no cover


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, rd.Dataset))
def _rename_ray_dataframe(df: rd.Dataset, columns: Dict[str, Any]) -> rd.Dataset:
    if len(columns) == 0:
        return df
    cols = _get_ray_dataframe_columns(df)
    missing = set(columns.keys()) - set(cols)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")
    new_cols = [columns.get(name, name) for name in cols]
    return df.map_batches(
        lambda b: b.rename_columns(new_cols), batch_format="pyarrow", **_ZERO_COPY
    )


def _get_arrow_tables(df: rd.Dataset) -> Iterable[pa.Table]:
    last_empty: Any = None
    empty = True
    for block in df.get_internal_block_refs():
        tb = ray.get(block)
        if tb.shape[0] > 0:
            yield tb
            empty = False
        else:
            last_empty = tb
    if empty:
        yield last_empty



import pyarrow as pa
from pytest import raises
from triad import Schema

import fugue.column.functions as f
from fugue.column import all_cols, col, lit, null


def test_is_agg():
    assert f.is_agg(f.first(col("a")))
    assert f.is_agg(f.count_distinct(col("a")).alias("x"))
    assert f.is_agg(f.first(col("a") + 1))
    assert f.is_agg(f.first(col("a")) + 1)
    assert f.is_agg((f.first(col("a")) < 1).alias("x"))
    assert f.is_agg(col("a") * f.first(col("a")) + 1)

    assert not f.is_agg(col("a"))
    assert not f.is_agg(lit("a"))
    assert not f.is_agg(col("a") + col("b"))
    assert not f.is_agg(null())


def test_functions():
    schema = Schema("a:int,b:str,c:bool,d:double")

    expr = f.coalesce(col("a"), 1, None, col("b") + col("c"))
    assert "COALESCE(a,1,NULL,+(b,c))" == str(expr)
    assert expr.infer_type(schema) is None

    expr = f.min(col("a"))
    assert "MIN(a)" == str(expr)
    assert pa.int32() == expr.infer_type(schema)
    assert "MIN(a) AS a" == str(expr.infer_alias())
    assert "CAST(MIN(a) AS long) AS a" == str(expr.cast(int).infer_alias())
    assert "MIN(a) AS b" == str(expr.alias("b").infer_alias())

    assert "MIN(-(a)) AS a" == str(f.min(-col("a")).infer_alias())

    expr = f.min(lit(1.1))
    assert "MIN(1.1)" == str(expr)
    assert pa.float64() == expr.infer_type(schema)

    expr = f.max(col("a"))
    assert "MAX(a)" == str(expr)
    assert pa.int32() == expr.infer_type(schema)

    expr = f.max(lit(1.1))
    assert "MAX(1.1)" == str(expr)
    assert pa.float64() == expr.infer_type(schema)

    expr = f.first(col("a"))
    assert "FIRST(a)" == str(expr)
    assert pa.int32() == expr.infer_type(schema)

    expr = f.first(lit(1.1))
    assert "FIRST(1.1)" == str(expr)
    assert pa.float64() == expr.infer_type(schema)

    expr = f.last(col("a"))
    assert "LAST(a)" == str(expr)
    assert pa.int32() == expr.infer_type(schema)

    expr = f.last(lit(1.1))
    assert "LAST(1.1)" == str(expr)
    assert pa.float64() == expr.infer_type(schema)

    expr = f.avg(col("a"))
    assert "AVG(a)" == str(expr)
    assert expr.infer_type(schema) is None

    expr = f.sum(col("a"))
    assert "SUM(a)" == str(expr)
    assert expr.infer_type(schema) is None

    expr = f.count(col("a"))
    assert "COUNT(a)" == str(expr)
    assert expr.infer_type(schema) is None

    expr = f.count_distinct(col("a"))
    assert "COUNT(DISTINCT a)" == str(expr)
    assert expr.infer_type(schema) is None
    assert "COUNT(DISTINCT a) AS a" == str(expr.infer_alias())

    expr = f.count_distinct(all_cols())
    assert "COUNT(DISTINCT *)" == str(expr)
    assert expr.infer_type(schema) is None
    raises(NotImplementedError, lambda: expr.infer_alias())



import fugue.api as fa
import pandas as pd
import polars as pl


def test_to_df():
    df = pl.from_pandas(pd.DataFrame({"a": [0, 1]}))
    res = fa.fugue_sql("SELECT * FROM df", df=df, engine="duckdb")
    assert fa.as_array(res) == [[0], [1]]

    df2 = pl.from_pandas(pd.DataFrame({"a": [0]}))
    res = fa.inner_join(df, df2, engine="duckdb")
    assert fa.as_array(res) == [[0]]



import pyspark
from fugue_spark._utils.convert import (
    to_cast_expression,
    to_schema,
    to_select_expression,
    to_spark_schema,
)
from pytest import raises


def test_schema_conversion(spark_session):
    def test(expr):
        assert to_schema(to_spark_schema(expr)) == expr

    test("a:int,b:long,c:[int],d:datetime,e:date,f:decimal(3,4),g:str")
    test("a:{a:[int],b:[str]}")
    test("a:[{a:int}]")
    s = to_spark_schema(to_spark_schema("a:int"))
    assert to_spark_schema(s) is s

    df = spark_session.createDataFrame([[1]], schema=to_spark_schema("a:int"))
    assert to_schema(to_spark_schema(df)) == "a:int"
    assert to_schema(df) == "a:int"
    assert to_schema(dict(a=str)) == "a:str"

    from pyspark.sql.types import (
        ArrayType,
        IntegerType,
        MapType,
        StringType,
        StructField,
        StructType,
    )

    schema = StructType(
        [
            StructField(
                "name",
                ArrayType(
                    StructType(
                        [
                            StructField("nest_name", StringType(), True),
                            StructField("nest_value", IntegerType(), True),
                        ]
                    ),
                    True,
                ),
                True,
            )
        ]
    )
    df = spark_session.createDataFrame([[[("a", 1), ("b", 2)]]], schema)
    assert to_schema(df) == "name:[{nest_name:str,nest_value:int}]"
    assert to_spark_schema("name:[{nest_name:str,nest_value:int}]") == schema

    if pyspark.__version__ >= "3":
        schema = StructType(
            [StructField("a", MapType(StringType(), IntegerType(), True), True)],
        )
        df = spark_session.createDataFrame([[{"x": 1}], [{"y": 2}]], schema)
        assert to_schema(df) == "a:<str,int>"
        assert to_spark_schema("a:<str,int>") == schema


def test_to_cast_expression():
    # length mismatch
    raises(ValueError, lambda: to_cast_expression("a:int,b:int", "a:int", False))
    assert (False, ["a", "b"]) == to_cast_expression(
        "a:int,b:int", "a:int,b:int", False
    )
    assert (False, ["a", "b"]) == to_cast_expression("a:int,b:int", "a:int,b:int", True)
    assert (True, ["aa AS a", "b"]) == to_cast_expression(
        "aa:int,b:int", "a:int,b:int", True
    )
    raises(ValueError, lambda: to_cast_expression("aa:int,b:int", "a:int,b:int", False))
    assert (True, ["CAST(a AS int) a", "b"]) == to_cast_expression(
        "a:long,b:int", "a:int,b:int", True
    )
    assert (True, ["CAST(aa AS int) a", "b"]) == to_cast_expression(
        "aa:long,b:int", "a:int,b:int", True
    )


def test_to_select_expression():
    assert to_select_expression("a:int,b:str", ["b", "a"]) == ["b", "a"]
    raises(KeyError, lambda: to_select_expression("a:int,b:str", ["b", "x"]))
    assert to_select_expression("a:int,b:str", "b:str,a:int") == ["b", "a"]
    assert to_select_expression("a:int,b:str", "b:str,a:long") == [
        "b",
        "CAST(a AS bigint) a",
    ]
    assert to_select_expression("a:int,b:double,c:float", "a:str,b:str,c:long") == [
        "CAST(a AS string) a",
        "CAST(IF(isnan(b) OR b IS NULL, NULL, b) AS string) b",
        "CAST(IF(isnan(c) OR c IS NULL, NULL, c) AS bigint) c",
    ]
    raises(KeyError, lambda: to_select_expression("a:int,b:str", "b:str,x:int"))



from datetime import datetime
from typing import Any

import pandas as pd
import pyspark
import pyspark.sql as ps
import pytest
from pyspark.sql import SparkSession
from triad.collections.schema import Schema

import fugue.api as fi
from fugue.dataframe.pandas_dataframe import PandasDataFrame
from fugue.plugins import get_column_names, rename
from fugue_spark import SparkExecutionEngine
from fugue_spark._utils.convert import to_schema, to_spark_schema
from fugue_spark.dataframe import SparkDataFrame
from fugue_test.dataframe_suite import DataFrameTests


class SparkDataFrameTests(DataFrameTests.Tests):
    @pytest.fixture(autouse=True)
    def init_session(self, spark_session):
        self.spark_session = spark_session

    def df(self, data: Any = None, schema: Any = None) -> SparkDataFrame:
        session = SparkSession.builder.getOrCreate()
        engine = SparkExecutionEngine(session)
        return engine.to_df(data, schema=schema)

    def test_alter_columns_invalid(self):
        # TODO: Spark will silently cast invalid data to nulls without exceptions
        pass

    def test_map_type(self):
        if pyspark.__version__ >= "3":
            return super().test_map_type()


class NativeSparkDataFrameTests(DataFrameTests.NativeTests):
    @pytest.fixture(autouse=True)
    def init_session(self, spark_session):
        self.spark_session = spark_session

    def df(self, data: Any = None, schema: Any = None):
        engine = SparkExecutionEngine(self.spark_session)
        return engine.to_df(data, schema=schema).native

    def to_native_df(self, pdf: pd.DataFrame) -> Any:
        return self.spark_session.createDataFrame(pdf)

    def test_not_local(self):
        assert not fi.is_local(self.df([], "a:int,b:str"))

    def test_alter_columns_invalid(self):
        # TODO: Spark will silently cast invalid data to nulls without exceptions
        pass

    def test_map_type(self):
        if pyspark.__version__ >= "3":
            return super().test_map_type()


def test_init(spark_session):
    sdf = spark_session.createDataFrame([["a", 1]])
    df = SparkDataFrame(sdf, "a:str,b:double")
    assert [["a", 1.0]] == df.as_array()
    assert [["a", 1.0]] == df.as_pandas().values.tolist()
    assert not df.is_local
    assert df.is_bounded
    assert df.num_partitions > 0

    df = _df([["a", 1], ["b", 2]])
    assert [["a", 1], ["b", 2]] == df.as_array()
    df = _df([], "a:str,b:str")
    assert [] == df.as_array()
    assert df.schema == "a:str,b:str"
    df = _df([["a", 1], ["b", 2]], "a:str,b:str")
    assert [["a", "1"], ["b", "2"]] == df.as_array()
    assert df.schema == "a:str,b:str"


def test_nested(spark_session):
    # data = [[dict(a=1, b=[3, 4], d=1.0)], [json.dumps(dict(b=[30, "40"]))]]
    # df = SparkDataFrame(data, "a:{a:str,b:[int]}")
    # a = df.as_array(type_safe=True)
    # assert [[dict(a="1", b=[3, 4])], [dict(a=None, b=[30, 40])]] == a

    data = [[[10, 20]]]
    sdf = spark_session.createDataFrame(data, to_spark_schema("a:[int]"))
    df = SparkDataFrame(sdf)
    assert data == df.as_array(type_safe=False)
    assert data == df.as_array(type_safe=True)
    assert data == list(df.as_array_iterable(type_safe=False))
    assert data == list(df.as_array_iterable(type_safe=True))

    data = [[dict(b=[30, 40])]]
    sdf = spark_session.createDataFrame(data, to_spark_schema("a:{a:str,b:[int]}"))
    df = SparkDataFrame(sdf)
    a = df.as_array(type_safe=False)
    assert [[dict(a=None, b=[30, 40])]] == a
    a = df.as_array(type_safe=True)
    assert [[dict(a=None, b=[30, 40])]] == a
    a = list(df.as_array_iterable(type_safe=False))
    assert [[dict(a=None, b=[30, 40])]] == a
    a = list(df.as_array_iterable(type_safe=True))
    assert [[dict(a=None, b=[30, 40])]] == a


def _test_as_array_perf():
    s = Schema()
    arr = []
    for i in range(100):
        s.append(f"a{i}:int")
        arr.append(i)
    for i in range(100):
        s.append(f"b{i}:int")
        arr.append(float(i))
    for i in range(100):
        s.append(f"c{i}:str")
        arr.append(str(i))
    data = []
    for i in range(5000):
        data.append(list(arr))
    df = SparkDataFrame(data, s)
    res = df.as_array()
    res = df.as_array(type_safe=True)
    nts, ts = 0.0, 0.0
    for i in range(10):
        t = datetime.now()
        res = df.as_array()
        nts += (datetime.now() - t).total_seconds()
        t = datetime.now()
        res = df.as_array(type_safe=True)
        ts += (datetime.now() - t).total_seconds()
    print(nts, ts)


def _df(data, schema=None):
    session = SparkSession.builder.getOrCreate()
    if schema is not None:
        pdf = PandasDataFrame(data, to_schema(schema))
        df = session.createDataFrame(pdf.native, to_spark_schema(schema))
    else:
        df = session.createDataFrame(data)
    return SparkDataFrame(df, schema)


def _test_get_column_names(spark_session):
    df = spark_session.createDataFrame(
        pd.DataFrame([[0, 1, 2]], columns=["0", "1", "2"])
    )
    assert get_column_names(df) == ["0", "1", "2"]


def _test_rename(spark_session):
    pdf = spark_session.createDataFrame(
        pd.DataFrame([[0, 1, 2]], columns=["a", "b", "c"])
    )
    df = rename(pdf, {})
    assert isinstance(df, ps.DataFrame)
    assert get_column_names(df) == ["a", "b", "c"]

    pdf = spark_session.createDataFrame(
        pd.DataFrame([[0, 1, 2]], columns=["0", "1", "2"])
    )
    df = rename(pdf, {"0": "_0", "1": "_1", "2": "_2"})
    assert isinstance(df, ps.DataFrame)
    assert get_column_names(df) == ["_0", "_1", "_2"]



from triad import to_uuid

from fugue.collections.sql import StructuredRawSQL, TempTableName, transpile_sql


@transpile_sql.candidate(
    lambda sql, from_dialect, to_dialect: from_dialect == "x" and to_dialect == "y"
)
def _dummy_transpile(sql, from_dialect, to_dialect):
    return sql.lower()


def test_transpile():
    assert 'SELECT "x" FROM tb' == transpile_sql(
        "SELECT `x` FROM tb", from_dialect="spark", to_dialect="duckdb"
    )
    assert "SELECT `x` FROM tb" == transpile_sql(
        "SELECT `x` FROM tb", from_dialect="spark", to_dialect=None
    )
    assert "SELECT `x` FROM tb" == transpile_sql(
        "SELECT `x` FROM tb", from_dialect=None, to_dialect=None
    )
    assert "SELECT `x` FROM tb" == transpile_sql(
        "SELECT `x` FROM tb", from_dialect=None, to_dialect="duckdb"
    )
    assert 'select "x" from tb' == transpile_sql(
        'SELECT "x" FROM tb', from_dialect="x", to_dialect="y"
    )


def test_parse_sql():
    def parse(sql):
        parts = StructuredRawSQL.from_expr(sql)._statements
        return "".join([p[1] if not p[0] else "!" + p[1] + "!" for p in parts])

    t1 = TempTableName()
    t2 = TempTableName()
    assert parse("") == ""
    assert parse(f"{t1}") == f"!{t1.key}!"
    assert parse(f" {t1} ") == f" !{t1.key}! "
    assert parse(f"SELECT * FROM {t1}") == f"SELECT * FROM !{t1.key}!"
    assert (
        parse(f"SELECT * FROM {t1} NATURAL JOIN {t2}")
        == f"SELECT * FROM !{t1.key}! NATURAL JOIN !{t2.key}!"
    )
    assert (
        parse(f"SELECT {t1}.* FROM {t1} NATURAL JOIN {t2} WHERE {t2}.x<1")
        == f"SELECT !{t1.key}!.* FROM !{t1.key}! "
        f"NATURAL JOIN !{t2.key}! WHERE !{t2.key}!.x<1"
    )

    q = StructuredRawSQL.from_expr("SELECT * FROM abc", dialect="y")
    assert q.dialect == "y"


def test_structured():
    s = [
        (False, "SELECT * FROM"),
        (True, "tb1"),
        (False, "NATURAL JOIN"),
        (True, "tb2"),
    ]
    q = StructuredRawSQL(s, dialect="x")
    assert q.dialect == "x"
    assert q.construct() == "SELECT * FROM tb1 NATURAL JOIN tb2"
    assert q.construct(dialect="y") == "select * from tb1 natural join tb2"
    assert (
        q.construct({"tb1": "tt"}, dialect="y") == "select * from tt natural join tb2"
    )
    assert (
        q.construct(lambda x: x + "_", dialect="y")
        == "select * from tb1_ natural join tb2_"
    )


def test_uid():
    s1 = [(False, "SELECT * FROM"), (True, "tb1")]
    s2 = [(False, "SELECT * FROM"), (True, "tb1")]
    s3 = [(False, "SELECT * from"), (True, "tb1")]
    assert to_uuid(StructuredRawSQL(s1)) == to_uuid(StructuredRawSQL(s1))
    assert to_uuid(StructuredRawSQL(s1)) != to_uuid(StructuredRawSQL(s1, dialect="x"))
    assert to_uuid(StructuredRawSQL(s1)) == to_uuid(StructuredRawSQL(s2))
    assert to_uuid(StructuredRawSQL(s1)) != to_uuid(StructuredRawSQL(s3))



from fugue.dataframe import DataFrames
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.pandas_dataframe import PandasDataFrame
from pytest import raises
from triad.exceptions import InvalidOperationError


def test_dataframes():
    df1 = ArrayDataFrame([[0]], "a:int")
    df2 = ArrayDataFrame([[1]], "a:int")
    dfs = DataFrames(a=df1, b=df2)
    assert dfs[0] is df1
    assert dfs[1] is df2

    dfs = DataFrames([df1, df2], df1)
    assert not dfs.has_key
    assert dfs[0] is df1
    assert dfs[1] is df2
    assert dfs[2] is df1

    dfs2 = DataFrames(dfs, dfs, df2)
    assert not dfs2.has_key
    assert dfs2[0] is df1
    assert dfs2[1] is df2
    assert dfs2[2] is df1
    assert dfs2[3] is df1
    assert dfs2[4] is df2
    assert dfs2[5] is df1
    assert dfs2[6] is df2

    dfs = DataFrames([("a", df1), ("b", df2)])
    assert dfs.has_key
    assert dfs[0] is df1
    assert dfs[1] is df2
    assert dfs["a"] is df1
    assert dfs["b"] is df2

    with raises(ValueError):
        dfs["c"] = 1

    with raises(ValueError):
        dfs2 = DataFrames(1)

    with raises(ValueError):
        dfs2 = DataFrames(a=df1, b=2)

    with raises(InvalidOperationError):
        dfs2 = DataFrames(dict(a=df1), df2)

    with raises(InvalidOperationError):
        dfs2 = DataFrames(df2, dict(a=df1))

    with raises(InvalidOperationError):
        dfs2 = DataFrames(df1, a=df2)

    with raises(InvalidOperationError):
        dfs2 = DataFrames(DataFrames(df1, df2), x=df2)

    dfs2 = DataFrames(dfs)
    assert dfs2.has_key
    assert dfs2[0] is df1
    assert dfs2[1] is df2

    dfs1 = DataFrames(a=df1, b=df2)
    dfs2 = dfs1.convert(lambda x: PandasDataFrame(x.as_array(), x.schema))
    assert len(dfs1) == len(dfs2)
    assert dfs2.has_key
    assert isinstance(dfs2["a"], PandasDataFrame)
    assert isinstance(dfs2["b"], PandasDataFrame)

    dfs1 = DataFrames(df1, df2)
    dfs2 = dfs1.convert(lambda x: PandasDataFrame(x.as_array(), x.schema))
    assert len(dfs1) == len(dfs2)
    assert not dfs2.has_key
    assert isinstance(dfs2[0], PandasDataFrame)
    assert isinstance(dfs2[1], PandasDataFrame)


from typing import Any, List

import pandas as pd
from pytest import raises

from fugue import (
    NativeExecutionEngine,
    make_execution_engine,
    make_sql_engine,
    register_default_execution_engine,
    register_default_sql_engine,
    register_execution_engine,
    register_sql_engine,
)
from fugue.constants import FUGUE_CONF_SQL_IGNORE_CASE
from fugue.exceptions import FuguePluginsRegistrationError
from fugue.execution.factory import (
    infer_execution_engine,
    is_pandas_or,
    make_execution_engine,
    make_sql_engine,
    register_default_execution_engine,
    register_default_sql_engine,
    register_execution_engine,
    register_sql_engine,
)
from fugue_duckdb import DuckDBEngine


class _MockExecutionEngine(NativeExecutionEngine):
    def __init__(self, conf: Any, other: int = 0):
        super().__init__(conf=conf)
        self.other = other


class Dummy1:
    pass


class Dummy2(Dummy1):
    pass


class _MockExecutionEngine2(NativeExecutionEngine):
    def __init__(self, obj: Any, conf: Any, other: int = 0):
        assert isinstance(obj, Dummy2)
        super().__init__(conf=conf)
        self.other = other


class _MockSQlEngine(DuckDBEngine):
    def __init__(self, execution_engine, other: int = 1):
        super().__init__(execution_engine)
        self.other = other


@infer_execution_engine.candidate(lambda objs: is_pandas_or(objs, Dummy2))
def _infer_dummy_engine(objs: List[Any]) -> Any:
    return _MockExecutionEngine2(Dummy2(), {})


def test_execution_engine_alias():
    assert isinstance(make_execution_engine(), NativeExecutionEngine)

    c = make_execution_engine(conf={"a": 2})
    assert 2 == c.conf.get_or_throw("a", int)

    c = make_execution_engine(_MockExecutionEngine(conf={"a": 3}, other=4))
    assert isinstance(c, _MockExecutionEngine)
    assert 3 == c.conf.get_or_throw("a", int)
    assert 4 == c.other

    raises(FuguePluginsRegistrationError, lambda: make_execution_engine("xyz"))

    register_execution_engine(
        "__xyz", lambda conf, **kwargs: _MockExecutionEngine(conf, **kwargs)
    )
    c = make_execution_engine("__xyz")
    assert isinstance(c, _MockExecutionEngine)

    raises(
        ValueError,
        lambda: register_execution_engine(
            "__xyz",
            lambda conf, **kwargs: _MockExecutionEngine(conf, **kwargs),
            on_dup="dummy",
        ),
    )

    c = make_execution_engine("__xyz", conf={"a": 3}, other=4)
    assert isinstance(c, _MockExecutionEngine)
    assert 3 == c.conf.get_or_throw("a", int)
    assert 4 == c.other

    assert isinstance(make_execution_engine(), NativeExecutionEngine)

    register_default_execution_engine(
        lambda conf, **kwargs: _MockExecutionEngine(conf, **kwargs)
    )
    assert isinstance(make_execution_engine(), _MockExecutionEngine)

    c = make_execution_engine(conf={"a": 3}, other=4)
    assert isinstance(c, _MockExecutionEngine)
    assert 3 == c.conf.get_or_throw("a", int)
    assert 4 == c.other

    # MUST HAVE THIS STEP, or other tests will fail
    _reset()


def test_execution_engine_type():
    register_execution_engine(
        Dummy1,
        lambda engine, conf, **kwargs: _MockExecutionEngine2(engine, conf, **kwargs),
    )
    e = make_execution_engine(Dummy2(), {"b": 1}, other=2)
    assert isinstance(e, _MockExecutionEngine2)
    assert 1 == e.conf.get_or_throw("b", int)
    assert 2 == e.other


def test_sql_engine():
    assert not isinstance(
        make_sql_engine(None, make_execution_engine()), _MockSQlEngine
    )
    assert isinstance(
        make_sql_engine(_MockSQlEngine, make_execution_engine()), _MockSQlEngine
    )

    with raises(FuguePluginsRegistrationError):
        make_sql_engine("__dummy_no")

    register_execution_engine("__a", lambda conf: _MockExecutionEngine(conf))
    register_sql_engine("__aq", lambda engine: _MockSQlEngine(engine, other=11))
    e = make_execution_engine(("__a", "__aq"))
    assert isinstance(e, _MockExecutionEngine)
    assert isinstance(e.sql_engine, _MockSQlEngine)
    assert 0 == e.other
    assert 11 == e.sql_engine.other

    register_default_execution_engine(lambda conf: _MockExecutionEngine(conf))
    e = make_execution_engine()
    assert isinstance(e, _MockExecutionEngine)
    assert not isinstance(e.sql_engine, _MockSQlEngine)
    register_default_sql_engine(lambda engine: _MockSQlEngine(engine))
    e = make_execution_engine()
    assert isinstance(e, _MockExecutionEngine)
    assert isinstance(e.sql_engine, _MockSQlEngine)

    # SQL Engine override
    e = make_execution_engine(NativeExecutionEngine)
    assert type(e) == NativeExecutionEngine
    assert isinstance(e.sql_engine, _MockSQlEngine)

    # conditional override
    def to_sql_engine(engine):
        if isinstance(engine, _MockExecutionEngine):
            return _MockSQlEngine(engine)
        else:
            return engine.sql_engine

    register_default_sql_engine(to_sql_engine)
    e = make_execution_engine(NativeExecutionEngine)
    assert type(e) == NativeExecutionEngine
    assert type(e.sql_engine) != _MockSQlEngine

    e = make_execution_engine(_MockExecutionEngine)
    assert type(e) == _MockExecutionEngine
    assert type(e.sql_engine) == _MockSQlEngine

    # MUST HAVE THIS STEP, or other tests will fail
    _reset()


def test_global_funcs():
    assert isinstance(make_execution_engine(), NativeExecutionEngine)
    register_execution_engine(
        "xyz", lambda conf, **kwargs: _MockExecutionEngine(conf, **kwargs)
    )
    assert isinstance(make_execution_engine("xyz"), _MockExecutionEngine)
    register_default_execution_engine(
        lambda conf, **kwargs: _MockExecutionEngine(conf, **kwargs), on_dup="ignore"
    )
    assert not isinstance(make_execution_engine(), _MockExecutionEngine)
    register_default_execution_engine(
        lambda conf, **kwargs: _MockExecutionEngine(conf, **kwargs), on_dup="overwrite"
    )
    assert isinstance(make_execution_engine(), _MockExecutionEngine)

    se = DuckDBEngine(make_execution_engine())
    assert make_sql_engine(se) is se
    assert not isinstance(
        make_sql_engine(None, make_execution_engine()), _MockSQlEngine
    )
    register_sql_engine("x", lambda engine: _MockSQlEngine(engine))
    assert isinstance(make_sql_engine("x", make_execution_engine()), _MockSQlEngine)
    register_default_sql_engine(lambda engine: _MockSQlEngine(engine, other=10))
    e = make_execution_engine()
    assert isinstance(e, _MockExecutionEngine)
    assert isinstance(e.sql_engine, _MockSQlEngine)
    assert 10 == e.sql_engine.other

    # MUST HAVE THIS STEP, or other tests will fail
    _reset()


def test_make_execution_engine():
    e = make_execution_engine(None, {FUGUE_CONF_SQL_IGNORE_CASE: True})
    assert isinstance(e, NativeExecutionEngine)
    assert e.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)

    e = make_execution_engine(NativeExecutionEngine, {FUGUE_CONF_SQL_IGNORE_CASE: True})
    assert isinstance(e, NativeExecutionEngine)
    assert e.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)

    e = make_execution_engine(
        NativeExecutionEngine({"ab": "c"}), {FUGUE_CONF_SQL_IGNORE_CASE: True}, de="f"
    )
    assert isinstance(e, NativeExecutionEngine)
    assert e.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)
    assert "c" == e.conf.get_or_throw("ab", str)
    assert "f" == e.conf.get_or_throw("de", str)
    assert "c" == e.conf.get_or_throw("ab", str)

    e = make_execution_engine("pandas", {FUGUE_CONF_SQL_IGNORE_CASE: True})
    assert isinstance(e, NativeExecutionEngine)
    assert e.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)

    e = make_execution_engine(
        (NativeExecutionEngine, "duckdb"), {FUGUE_CONF_SQL_IGNORE_CASE: True}
    )
    assert isinstance(e, NativeExecutionEngine)
    assert e.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)

    e = make_execution_engine(NativeExecutionEngine({FUGUE_CONF_SQL_IGNORE_CASE: True}))
    assert isinstance(e, NativeExecutionEngine)
    assert e.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)

    e = make_execution_engine(
        (NativeExecutionEngine({FUGUE_CONF_SQL_IGNORE_CASE: True}), "duckdb")
    )
    e = make_execution_engine(e)
    assert isinstance(e, NativeExecutionEngine)
    assert e.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool)
    assert isinstance(e.sql_engine, DuckDBEngine)

    assert isinstance(make_sql_engine(None, e), DuckDBEngine)

    # MUST HAVE THIS STEP, or other tests will fail
    _reset()


def test_context_and_infer_execution_engine():
    e1 = _MockExecutionEngine({})
    e2 = _MockExecutionEngine2(Dummy2(), {})
    assert not e1.in_context and not e2.in_context
    with e2.as_context():
        assert not e1.in_context and e2.in_context
        with e1.as_context() as ex:
            assert e1.in_context and e2.in_context
            assert ex is e1
            e = make_execution_engine(
                None, conf={"x": False}, infer_by=[pd.DataFrame(), Dummy2()]
            )
            assert isinstance(e, _MockExecutionEngine)
            assert not isinstance(e, _MockExecutionEngine2)
            assert not e.conf["x"]
        assert not e1.in_context and e2.in_context

        e = make_execution_engine(None, conf={"x": True})
        assert isinstance(e, _MockExecutionEngine2)

    assert not e1.in_context and not e2.in_context

    e = make_execution_engine(None)
    assert isinstance(e, NativeExecutionEngine)
    assert not isinstance(e, _MockExecutionEngine)

    e = make_execution_engine(
        None, conf={"x": True}, infer_by=[pd.DataFrame(), Dummy2()]
    )
    assert isinstance(e, _MockExecutionEngine2)
    assert e.conf["x"]


def test_is_pandas_or():
    assert not is_pandas_or([], int)
    assert is_pandas_or([1], int)
    assert not is_pandas_or([pd.DataFrame()], int)
    assert is_pandas_or([1, pd.DataFrame()], int)
    assert is_pandas_or([1, pd.DataFrame()], (int, bool))
    assert is_pandas_or([1, pd.DataFrame(), False], (int, bool))
    assert not is_pandas_or([pd.DataFrame(), pd.DataFrame()], (int, bool))


def _reset():
    register_default_execution_engine(
        lambda conf, **kwargs: NativeExecutionEngine(conf)
    )
    register_default_sql_engine(lambda ee, **kwargs: ee.sql_engine)



import dask.dataframe as dd
import pandas as pd
from fugue import register_execution_engine
from fugue import FugueSQLWorkflow

from fugue_dask import DaskExecutionEngine


def test_sql():
    register_execution_engine(
        "da", lambda conf, **kwargs: DaskExecutionEngine(conf=conf)
    )
    df = dd.from_pandas(pd.DataFrame([[0], [1]], columns=["a"]), npartitions=2)
    dag = FugueSQLWorkflow()
    dag(
        """
    SELECT * FROM df WHERE a>0
    PRINT
    """,
        df=df,
    )
    dag.run("da")



from typing import Dict, Any
import pyspark

FUGUE_SPARK_CONF_USE_PANDAS_UDF = "fugue.spark.use_pandas_udf"

FUGUE_SPARK_DEFAULT_CONF: Dict[str, Any] = {FUGUE_SPARK_CONF_USE_PANDAS_UDF: True}

_IS_SPARK_2 = pyspark.__version__ < "3"



import pickle
from threading import RLock
from typing import Any, List, Optional

import dask.dataframe as dd
import pandas as pd
from dask.distributed import Client
import dask

import fugue.api as fa
from fugue import transform
from fugue.collections.partition import PartitionSpec
from fugue.dataframe.pandas_dataframe import PandasDataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.plugins import infer_execution_engine
from fugue.workflow.workflow import FugueWorkflow
from fugue_dask.dataframe import DaskDataFrame
from fugue_dask.execution_engine import DaskExecutionEngine
from fugue_test.builtin_suite import BuiltInTests
from fugue_test.execution_suite import ExecutionEngineTests

_CONF = {
    "fugue.rpc.server": "fugue.rpc.flask.FlaskRPCServer",
    "fugue.rpc.flask_server.host": "127.0.0.1",
    "fugue.rpc.flask_server.port": "1234",
    "fugue.rpc.flask_server.timeout": "2 sec",
}


class DaskExecutionEngineTests(ExecutionEngineTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._engine = cls.make_engine(cls)
        fa.set_global_engine(cls._engine)

    @classmethod
    def tearDownClass(cls):
        fa.clear_global_engine()
        cls._engine.dask_client.close()

    def make_engine(self):
        client = Client(processes=True, n_workers=3, threads_per_worker=1)
        # p2p (new default algo has bugs)
        dask.config.set({"dataframe.shuffle.method": "tasks"})
        e = DaskExecutionEngine(client, conf=dict(test=True, **_CONF))
        return e

    def test_properties(self):
        assert self.engine.is_distributed
        assert self.engine.map_engine.is_distributed
        assert self.engine.sql_engine.is_distributed

    def test_get_parallelism(self):
        assert fa.get_current_parallelism() == 3

    def test__join_outer_pandas_incompatible(self):
        return

    def test_to_df(self):
        e = self.engine
        a = e.to_df([[1, 2], [3, 4]], "a:int,b:int")
        df_eq(a, [[1, 2], [3, 4]], "a:int,b:int", throw=True)
        a = e.to_df(PandasDataFrame([[1, 2], [3, 4]], "a:int,b:int"))
        df_eq(a, [[1, 2], [3, 4]], "a:int,b:int", throw=True)
        assert a is e.to_df(a)

    def test_repartition(self):
        e = self.engine
        a = e.to_df([[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]], "a:int,b:int")
        b = e.repartition(a, PartitionSpec())
        assert a is b
        b = e.repartition(a, PartitionSpec(num=3))
        assert 3 == b.num_partitions
        b = e.repartition(a, PartitionSpec(num="0"))
        assert a is b
        b = e.repartition(a, PartitionSpec(num="ROWCOUNT"))
        assert 5 == b.num_partitions
        b = e.repartition(a, PartitionSpec(num="ROWCOUNT/2"))
        assert 2 == b.num_partitions
        b = e.repartition(a, PartitionSpec(num="ROWCOUNT-ROWCOUNT"))
        assert a is b
        b = e.repartition(a, PartitionSpec(by=["a"], num=3))
        assert a.num_partitions == b.num_partitions

        b = e.repartition(a, PartitionSpec(num=5, algo="even"))
        assert 5 == b.num_partitions
        for i in range(b.num_partitions):
            assert len(b.native.partitions[i]) == 1

        b = e.repartition(a, PartitionSpec(num="ROWCOUNT", algo="even"))
        assert 5 == b.num_partitions
        for i in range(b.num_partitions):
            assert len(b.native.partitions[i]) == 1

        b = e.repartition(a, PartitionSpec(num="ROWCOUNT*2", algo="even"))
        assert 5 == b.num_partitions
        for i in range(b.num_partitions):
            assert len(b.native.partitions[i]) == 1

        b = e.repartition(a, PartitionSpec(num="ROWCOUNT-4", algo="even"))
        assert 1 == b.num_partitions
        for i in range(b.num_partitions):
            assert len(b.native.partitions[i]) == 5

        b = e.repartition(a, PartitionSpec(num="ROWCOUNT-ROWCOUNT", algo="even"))
        assert a is b

    def test_sample_n(self):
        # TODO: dask does not support sample by number of rows
        pass

    def test_infer_engine(self):
        ddf = dd.from_pandas(pd.DataFrame([[0]], columns=["a"]), npartitions=2)
        assert isinstance(infer_execution_engine([ddf]), Client)

        fdf = DaskDataFrame(ddf)
        assert isinstance(infer_execution_engine([fdf]), Client)


class DaskExecutionEngineBuiltInTests(BuiltInTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._engine = cls.make_engine(cls)

    @classmethod
    def tearDownClass(cls):
        cls._engine.dask_client.close()

    def make_engine(self):
        e = DaskExecutionEngine(conf=dict(test=True, **_CONF))
        return e

    def test_yield_table(self):
        pass

    def test_default_init(self):
        a = FugueWorkflow().df([[0]], "a:int")
        df_eq(a.compute(DaskExecutionEngine), [[0]], "a:int")

    def test_annotation(self):
        def m_c(engine: DaskExecutionEngine) -> dd.DataFrame:
            return dd.from_pandas(pd.DataFrame([[0]], columns=["a"]), npartitions=2)

        def m_p(engine: DaskExecutionEngine, df: dd.DataFrame) -> dd.DataFrame:
            return df

        def m_o(engine: DaskExecutionEngine, df: dd.DataFrame) -> None:
            assert 1 == df.compute().shape[0]

        with FugueWorkflow() as dag:
            df = dag.create(m_c).process(m_p)
            df.assert_eq(dag.df([[0]], "a:long"))
            df.output(m_o)
        dag.run(self.engine)

    def test_coarse_partition(self):
        def verify_coarse_partition(df: pd.DataFrame) -> List[List[Any]]:
            ct = df.a.nunique()
            s = df.a * 1000 + df.b
            ordered = ((s - s.shift(1)).dropna() >= 0).all(axis=None)
            return [[ct, ordered]]

        def assert_(df: pd.DataFrame, rc: int, n: int, check_ordered: bool) -> None:
            if rc > 0:
                assert len(df) == rc
            assert df.ct.sum() == n
            if check_ordered:
                assert (df.ordered == True).all()

        gps = 100
        partition_num = 6
        df = pd.DataFrame(dict(a=list(range(gps)) * 10, b=range(gps * 10))).sample(
            frac=1.0
        )
        with FugueWorkflow() as dag:
            a = dag.df(df)
            c = a.partition(
                algo="coarse", by="a", presort="b", num=partition_num
            ).transform(verify_coarse_partition, schema="ct:int,ordered:bool")
            dag.output(
                c,
                using=assert_,
                params=dict(rc=partition_num, n=gps, check_ordered=True),
            )
        dag.run(self.engine)


def test_transform():
    class CB:
        def __init__(self):
            self._lock = RLock()
            self.n = 0

        def add(self, n):
            with self._lock:
                self.n += n

    cb = CB()

    def tr(df: List[List[Any]], add: Optional[callable]) -> List[List[Any]]:
        if add is not None:
            add(len(df))
        return [[pickle.dumps(x[0])] for x in df]

    pdf = pd.DataFrame(dict(a=list(range(5))))
    res = transform(
        pdf,
        tr,
        schema="b:binary",
        callback=cb.add,
        as_local=True,
        as_fugue=True,
        engine="dask",
        engine_conf=_CONF,
    )
    assert res.is_local
    assert 5 == res.count()
    assert 5 == cb.n

    res = transform(
        pdf,
        tr,
        schema="b:binary",
        as_fugue=True,
        engine="dask",
    )
    assert not res.is_local
    assert 5 == res.count()

    cb = CB()

    res = transform(
        pdf,
        tr,
        schema="b:binary",
        callback=cb.add,
        as_fugue=True,
        engine="dask",
        engine_conf=_CONF,
        persist=True,  # when you have a persist, you can use callback
    )
    assert not res.is_local
    assert 5 == res.count()
    assert 5 == cb.n



import pytest


def _is_spark2():
    try:
        import pyspark

        return pyspark.__version__ < "3.0.0"
    except Exception:  # pragma: no cover
        return False


skip_spark2 = pytest.mark.skipif(_is_spark2(), reason="Skip Spark<3")



import sys
from datetime import datetime
from typing import Any

import ibis
import pandas as pd
import pyarrow as pa
import pytest

import fugue.api as fe
from fugue import ArrowDataFrame
from fugue_duckdb.dataframe import DuckDataFrame
from fugue_test.dataframe_suite import DataFrameTests

from .mock.dataframe import MockDuckDataFrame


@pytest.mark.skipif(sys.version_info < (3, 8), reason="< 3.8")
class IbisDataFrameTests(DataFrameTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._con = ibis.duckdb.connect()

    def df(self, data: Any = None, schema: Any = None) -> DuckDataFrame:
        df = ArrowDataFrame(data, schema)
        name = f"_{id(df.native)}"
        # self._con.con.execute("register", (name, df.native))
        self._con.register(df.native, name)
        return MockDuckDataFrame(self._con.table(name), schema=schema)

    def test_init_df(self):
        df = self.df([["x", 1]], "a:str,b:int")
        df = MockDuckDataFrame(df.native, "a:str,b:long")
        assert df.schema == "a:str,b:long"

    def test_is_local(self):
        df = self.df([["x", 1]], "a:str,b:int")
        assert not fe.is_local(df)
        assert fe.is_bounded(df)

    def test_map_type(self):
        pass

    def test_as_arrow(self):
        # empty
        df = self.df([], "a:int,b:int")
        assert [] == list(ArrowDataFrame(df.as_arrow()).as_dict_iterable())
        # pd.Nat
        df = self.df([[pd.NaT, 1]], "a:datetime,b:int")
        assert [dict(a=None, b=1)] == list(
            ArrowDataFrame(df.as_arrow()).as_dict_iterable()
        )
        # pandas timestamps
        df = self.df([[pd.Timestamp("2020-01-01"), 1]], "a:datetime,b:int")
        assert [dict(a=datetime(2020, 1, 1), b=1)] == list(
            ArrowDataFrame(df.as_arrow()).as_dict_iterable()
        )

    def test_deep_nested_types(self):
        pass

    def test_list_type(self):
        pass


@pytest.mark.skipif(sys.version_info < (3, 8), reason="< 3.8")
class NativeIbisDataFrameTests(DataFrameTests.NativeTests):
    @classmethod
    def setUpClass(cls):
        cls._con = ibis.duckdb.connect()

    def df(self, data: Any = None, schema: Any = None):
        df = ArrowDataFrame(data, schema)
        name = f"_{id(df.native)}"
        # self._con.con.execute("register", (name, df.native))
        self._con.register(df.native, name)
        return MockDuckDataFrame(self._con.table(name), schema=schema).native

    def to_native_df(self, pdf: pd.DataFrame) -> Any:
        name = f"_{id(pdf)}"
        # self._con.con.execute("register", (name, pa.Table.from_pandas(pdf)))
        self._con.register(pa.Table.from_pandas(pdf), name)
        return self._con.table(name)

    def test_is_local(self):
        df = self.df([["x", 1]], "a:str,b:int")
        assert not fe.is_local(df)
        assert fe.is_bounded(df)

    def test_map_type(self):
        pass

    def test_as_arrow(self):
        pass

    def test_deep_nested_types(self):
        pass

    def test_list_type(self):
        pass



from pytest import raises
from triad import Schema, to_uuid

import fugue.column.functions as f
from fugue.column import (
    SelectColumns,
    SQLExpressionGenerator,
    all_cols,
    col,
    function,
    lit,
    null,
)
from fugue.column.expressions import _BinaryOpExpr


def test_select_columns():
    # not all with names
    cols = SelectColumns(
        col("a"), lit(1, "b"), col("bb") + col("cc"), f.first(col("c"))
    )
    assert to_uuid(cols) == to_uuid(cols)
    raises(ValueError, lambda: cols.assert_all_with_names())

    # distinct
    cols2 = SelectColumns(
        col("a"),
        lit(1, "b"),
        col("bb") + col("cc"),
        f.first(col("c")),
        arg_distinct=True,
    )
    assert to_uuid(cols) != to_uuid(cols2)

    # duplicated names
    cols = SelectColumns(col("a").alias("b"), lit(1, "b"))
    assert to_uuid(cols) != to_uuid(SelectColumns(col("a").alias("b"), lit(1, "c")))
    raises(ValueError, lambda: cols.assert_all_with_names())

    # with *, all cols must have alias
    cols = SelectColumns(all_cols(), col("a")).assert_no_agg()
    raises(ValueError, lambda: cols.assert_all_with_names())

    # * can be used at most once
    raises(
        ValueError, lambda: SelectColumns(all_cols(), all_cols(), col("a").alias("p"))
    )

    # * can't be used with aggregation
    raises(ValueError, lambda: SelectColumns(all_cols(), f.first(col("a")).alias("x")))

    cols = SelectColumns(
        col("aa").alias("a").cast(int),
        lit(1, "b"),
        (col("bb") + col("cc")).alias("c"),
        f.first(col("c")).alias("d"),
    ).assert_all_with_names()
    raises(AssertionError, lambda: cols.assert_no_agg())
    assert not cols.simple
    assert 1 == len(cols.simple_cols)
    assert "CAST(aa AS long) AS a" == str(cols.simple_cols[0])
    assert cols.has_literals
    assert 1 == len(cols.literals)
    assert "1 AS b" == str(cols.literals[0])
    assert cols.has_agg
    assert 1 == len(cols.non_agg_funcs)
    assert "+(bb,cc) AS c" == str(cols.non_agg_funcs[0])
    assert 1 == len(cols.agg_funcs)
    assert "FIRST(c) AS d" == str(cols.agg_funcs[0])
    assert 2 == len(cols.group_keys)  # a, c
    assert "aa" == cols.group_keys[0].output_name
    assert "" == cols.group_keys[1].output_name
    assert isinstance(cols.group_keys[1], _BinaryOpExpr)

    cols = SelectColumns(col("a")).assert_no_wildcard()
    assert cols.simple
    assert not cols.has_literals
    assert not cols.has_agg

    cols = SelectColumns(col("x"), all_cols(), col("y") + col("z"))
    cols = cols.replace_wildcard(Schema("a:int,b:int"))
    assert "x" == str(cols.all_cols[0])


def test_basic():
    gen = SQLExpressionGenerator()
    assert "a" == gen.generate(col("a"))
    assert "a AS bc" == gen.generate(col("a").alias("bc"))

    assert "'a'" == gen.generate(lit("a"))
    assert "'a' AS bc" == gen.generate(lit("a").alias("bc"))

    assert "CAST(a AS long) AS a" == gen.generate(col("a").cast(int))


def test_select_exprs():
    gen = SQLExpressionGenerator()
    assert "(a+2)*3" == gen.generate((col("a") + 2) * 3)
    assert "(-a+2)*3" == gen.generate((-col("a") + 2) * 3)
    assert "(a*2)/3 AS x" == gen.generate(((col("a") * 2) / 3).alias("x"))
    assert "COUNT(DISTINCT a) AS x" == gen.generate(
        (f.count_distinct(col("a"))).alias("x")
    )


def test_conditions():
    gen = SQLExpressionGenerator()
    assert "(a=-1) AND (b>=c)" == gen.generate(
        (col("a") == -1) & (col("b") >= col("c"))
    )
    assert "TRUE AND (b>=c)" == gen.generate(True & (col("b") >= col("c")))
    assert "TRUE AND NOT (b>=c)" == gen.generate(True & ~(col("b") >= col("c")))
    assert "TRUE OR (b>=c) IS NOT NULL" == gen.generate(
        True | (col("b") >= col("c")).not_null()
    )


def test_functions():
    gen = SQLExpressionGenerator()
    assert "COALESCE(a,b+c,(d+e)-1,NULL) IS NULL" == gen.generate(
        f.coalesce(
            col("a"), col("b") + col("c"), col("d") + col("e") - 1, null()
        ).is_null()
    )
    assert (
        "MY(MIN(`x y`),MAX(y+1),AVG(z),2,aa=FIRST(a),bb=LAST('b'),cc=COUNT(DISTINCT *)) AS `x z`"
        == gen.generate(
            function(
                "MY",
                f.min(col("x y")),
                f.max(col("y") + 1),
                f.avg(col("z")),
                2,
                aa=f.first(col("a")),
                bb=f.last(lit("b")),
                cc=f.count_distinct(all_cols()),
            ).alias("x z")
        )
    )

    def dummy(expr):
        yield "DUMMY"
        if expr.is_distinct:
            yield " D"

    gen.add_func_handler("MY", dummy)
    assert "DUMMY D AS x" == gen.generate(
        function("MY", 2, 3, arg_distinct=True).alias("x")
    )


def test_where():
    gen = SQLExpressionGenerator()
    assert "SELECT * FROM !x! WHERE (`a `<5) AND b IS NULL" == _to_sql(
        gen.where((col("a ") < 5) & col("b").is_null(), "x")
    )
    assert "SELECT * FROM !x! WHERE a<5" == _to_sql(
        gen.where((col("a") < 5).alias("x"), "x")
    )
    raises(ValueError, lambda: list(gen.where(f.max(col("a")), "x")))


def test_select():
    gen = SQLExpressionGenerator()

    # no aggregation
    cols = SelectColumns(all_cols())
    assert "SELECT * FROM !x!" == _to_sql(gen.select(cols, "x"))

    cols = SelectColumns(col("a b"), lit(1).alias("b"), (col("b") + col("c")).alias("x"))
    where = (col("a b") > 5).alias("aa")
    assert "SELECT `a b`, 1 AS b, b+c AS x FROM !t! WHERE `a b`>5" == _to_sql(
        gen.select(cols, "t", where=where)
    )

    # aggregation without literals
    cols = SelectColumns(f.max(col("c")).alias("c"), col("a", "aa"), col("b"))
    assert "SELECT MAX(c) AS c, a AS aa, b FROM !t! GROUP BY a, b" == _to_sql(
        gen.select(cols, "t")
    )

    where = col("a") < 10
    having = (f.max(col("a")) > 5).alias("aaa")
    assert (
        "SELECT MAX(c) AS c, a AS aa, b FROM !t! WHERE a<10 GROUP BY a, b HAVING MAX(a)>5"
        == _to_sql(gen.select(cols, "t", where=where, having=having))
    )

    cols = SelectColumns(
        f.min(col("c") + 1).alias("c"),
        f.avg(col("d") + col("e")).cast(int).alias("d"),
    )
    assert "SELECT MIN(c+1) AS c, CAST(AVG(d+e) AS long) AS d FROM !t!" == _to_sql(
        gen.select(cols, "t")
    )

    # aggregation with literals
    cols = SelectColumns(
        lit(1, "k"), f.max(col("c")).alias("c"), lit(2, "j"), col("a", "aa"), col("b")
    )
    assert (
        "SELECT 1 AS k, c, 2 AS j, aa, b FROM ( SELECT MAX(c) AS c, a AS aa, b FROM !t! GROUP BY a, b )"
        == _to_sql(gen.select(cols, "t"))
    )

    cols = SelectColumns(lit(1, "k"), f.max(col("c")).alias("c"), lit(2, "j"))
    assert "SELECT 1 AS k, c, 2 AS j FROM ( SELECT MAX(c) AS c FROM !t! )" == _to_sql(
        gen.select(cols, "t")
    )

    cols = SelectColumns(lit(1, "k"), col("a"), f.max(col("c")).alias("c"), lit(2, "j"))
    assert (
        "SELECT 1 AS k, a, c, 2 AS j FROM ( SELECT a, MAX(c) AS c FROM !t! GROUP BY a )"
        == _to_sql(gen.select(cols, "t"))
    )

    # cast
    cols = SelectColumns(
        col("c").cast(float),
        f.avg(col("d") + col("e")).cast(int).alias("d"),
    )
    assert (
        "SELECT CAST(c AS double) AS c, CAST(AVG(d+e) AS long) AS d FROM !t! GROUP BY c"
        == _to_sql(gen.select(cols, "t"))
    )

    # infer alias
    cols = SelectColumns(
        (-col("c")).cast(float),
        f.max(col("e")).cast(int),
        f.avg(col("d") + col("e")).cast(int).alias("d"),
    )
    assert (
        "SELECT CAST(-c AS double) AS c, CAST(MAX(e) AS long) AS e, "
        "CAST(AVG(d+e) AS long) AS d FROM !t! GROUP BY -c"
        == _to_sql(gen.select(cols, "t"))
    )


def test_correct_select_schema():
    schema = Schema("a:double,b:str")
    gen = SQLExpressionGenerator()

    sc = SelectColumns(all_cols(), col("c"))
    output = Schema("a:double,b:str,c:str")
    c = gen.correct_select_schema(schema, sc, output)
    assert c is None

    output = Schema("a:int,b:int,c:str")
    c = gen.correct_select_schema(schema, sc, output)
    assert c == "a:double,b:str"

    sc = SelectColumns(f.count(all_cols()).alias("t"), col("c").alias("a"))
    output = Schema("t:int,a:str")
    c = gen.correct_select_schema(schema, sc, output)
    assert c is None

    sc = SelectColumns((col("a") + col("b")).cast(str).alias("a"), lit(1, "c"))
    output = Schema("a:int,c:str")
    c = gen.correct_select_schema(schema, sc, output)
    assert c == "a:str,c:long"


def test_no_cast():
    gen = SQLExpressionGenerator(enable_cast=False)
    cols = SelectColumns(
        f.max(col("c")).cast("long").alias("c"), col("a", "aa"), col("b")
    )
    assert "SELECT MAX(c) AS c, a AS aa, b FROM !t! GROUP BY a, b" == _to_sql(
        gen.select(cols, "t")
    )


def _to_sql(parts):
    return (
        " ".join(
            "!" + x[1].strip() + "!" if x[0] else x[1].strip()
            for x in parts
            if x[1].strip() != ""
        )
    ).strip()



from pytest import raises

from fugue.collections.partition import PartitionSpec
from fugue.exceptions import (
    FugueWorkflowCompileValidationError,
    FugueWorkflowRuntimeValidationError,
)
from fugue.extensions._utils import (
    parse_validation_rules_from_comment,
    to_validation_rules,
    validate_input_schema,
    validate_partition_spec,
)


def test_parse_output_schema_from_comment():
    # partitionby_has: a, b
    # partitionby_is: c , d
    # presort_has: a, b desc
    # presort_is: a asc, b desc
    # input_has: a,b: str,c
    # input_is: a: int , b : str
    # dummy: ssdsfd
    def t():
        pass

    # input_is:
    def t2():
        pass

    assert {
        "partitionby_has": ["a", "b"],
        "partitionby_is": ["c", "d"],
        "presort_has": [("a", True), ("b", False)],
        "presort_is": [("a", True), ("b", False)],
        "input_has": ["a", "b:str", "c"],
        "input_is": "a:int,b:str",
    } == parse_validation_rules_from_comment(t)

    raises(SyntaxError, lambda: parse_validation_rules_from_comment(t2))


def test_to_validation_rules():
    raises(NotImplementedError, lambda: to_validation_rules({"": ""}))
    assert {
        "partitionby_has": ["a", "b"],
        "partitionby_is": ["c", "d"],
    } == to_validation_rules({"partitionby_has": ["a", "b"], "partitionby_is": " c, d"})
    assert {
        "presort_has": [("a", True), ("b", False)],
        "presort_is": [("c", False), ("d", True)],
    } == to_validation_rules(
        {
            "presort_has": " a , b desc ",
            "presort_is": [("c", False), "d"],
        }
    )
    assert {"input_has": ["a", "b:str"],} == to_validation_rules(
        {
            "input_has": ["a", " b : str "],
        }
    )
    assert {"input_has": ["a", "b:str"],} == to_validation_rules(
        {
            "input_has": "a, b : str",
        }
    )
    assert {"input_is": "a:int,b:str"} == to_validation_rules(
        {
            "input_is": ["a:int", " b : str "],
        }
    )
    with raises(SyntaxError):
        to_validation_rules(
            {
                "input_is": "a, b : str",
            }
        )


def test_validate_partition_spec():
    spec = PartitionSpec(by=["a", "b", "c"], presort="d desc, e")

    validate_partition_spec(spec, to_validation_rules({"partitionby_has": ["a"]}))
    with raises(FugueWorkflowCompileValidationError):
        validate_partition_spec(spec, to_validation_rules({"partitionby_has": ["x"]}))
    with raises(FugueWorkflowCompileValidationError):
        validate_partition_spec(spec, to_validation_rules({"partitionby_has": "a,x"}))
    validate_partition_spec(spec, to_validation_rules({"partitionby_is": "a,b,c"}))
    # order in partitionby doesn't matter
    validate_partition_spec(spec, to_validation_rules({"partitionby_is": "c,b,a"}))
    with raises(FugueWorkflowCompileValidationError):
        validate_partition_spec(spec, to_validation_rules({"partitionby_is": "a,b"}))

    # order in presort_has doesn't matter
    validate_partition_spec(spec, to_validation_rules({"presort_has": "d desc"}))
    validate_partition_spec(spec, to_validation_rules({"presort_has": "e,d desc"}))
    with raises(FugueWorkflowCompileValidationError):  # sort order matters
        validate_partition_spec(spec, to_validation_rules({"presort_has": "d"}))
    with raises(FugueWorkflowCompileValidationError):
        validate_partition_spec(spec, to_validation_rules({"presort_has": "x"}))
    with raises(FugueWorkflowCompileValidationError):
        validate_partition_spec(spec, to_validation_rules({"presort_has": "x,d desc"}))

    validate_partition_spec(spec, to_validation_rules({"presort_is": "d desc,e"}))
    with raises(FugueWorkflowCompileValidationError):  # order in presort_is matters
        validate_partition_spec(spec, to_validation_rules({"presort_is": "e,d desc"}))


def test_validate_input_schema():
    schema = "a:int,b:int,c:str"

    validate_input_schema(schema, to_validation_rules({"input_has": "a, c"}))
    validate_input_schema(schema, to_validation_rules({"input_has": "a, c,b:int"}))
    with raises(FugueWorkflowRuntimeValidationError):  # col type matters (if specified)
        validate_input_schema(schema, to_validation_rules({"input_has": "a, c,b:str"}))

    validate_input_schema(
        schema, to_validation_rules({"input_is": "a:int, b:int, c:str"})
    )
    with raises(FugueWorkflowRuntimeValidationError):  # col type matters (if specified)
        validate_input_schema(
            schema, to_validation_rules({"input_is": "a:int, b:int, c:int"})
        )
    with raises(FugueWorkflowRuntimeValidationError):  # order matters
        validate_input_schema(
            schema, to_validation_rules({"input_is": "a:int, c:str,b:int"})
        )
    with raises(SyntaxError):  # order matters
        validate_input_schema(schema, to_validation_rules({"input_is": "a,b,c"}))



from typing import Any, Dict, Iterable, List

import pandas as pd
from fugue.dataframe import ArrayDataFrame
from fugue.exceptions import FugueInterfacelessError
from fugue.extensions.transformer import (
    Transformer,
    _to_transformer,
    parse_transformer,
    register_transformer,
    transformer,
)
from pytest import raises
from triad import ParamDict, to_uuid
from triad.collections.schema import Schema


def test_transformer():
    assert isinstance(t1, Transformer)
    df = ArrayDataFrame([[0]], "a:int")
    t1._output_schema = t1.get_output_schema(df)
    assert t1.output_schema == "a:int,b:int"
    t2._output_schema = t2.get_output_schema(df)
    assert t2.output_schema == "b:int,a:int"
    t3._output_schema = t3.get_output_schema(df)
    assert t3.output_schema == "a:int,b:int"
    assert [[0, 1]] == list(t3(df.as_array_iterable()))


def test__to_transformer():
    a = _to_transformer(MockTransformer)
    assert isinstance(a, MockTransformer)
    b = _to_transformer("MockTransformer")
    assert isinstance(b, MockTransformer)

    a = _to_transformer(t1, None)
    assert isinstance(a, Transformer)
    a._x = 1
    # every parse should produce a different transformer even the input is
    # a transformer instance
    b = _to_transformer(t1, None)
    assert isinstance(b, Transformer)
    assert "_x" not in b.__dict__
    c = _to_transformer("t1", None)
    assert isinstance(c, Transformer)
    assert "_x" not in c.__dict__
    c._x = 1
    d = _to_transformer("t1", None)
    assert isinstance(d, Transformer)
    assert "_x" not in d.__dict__
    raises(FugueInterfacelessError, lambda: _to_transformer(t4, None))
    raises(FugueInterfacelessError, lambda: _to_transformer("t4", None))
    e = _to_transformer("t4", "*,b:int")
    assert isinstance(e, Transformer)
    f = _to_transformer("t5")
    assert isinstance(f, Transformer)
    g = _to_transformer("t6", "*,b:int")
    assert isinstance(g, Transformer)
    assert g.get_format_hint() is None
    h = _to_transformer("t7")
    assert isinstance(h, Transformer)
    assert h.get_format_hint() == "pandas"
    i = _to_transformer("t8")
    assert isinstance(i, Transformer)
    j = _to_transformer("t9")
    assert isinstance(j, Transformer)
    k = _to_transformer("t10")
    assert isinstance(k, Transformer)


def test__register():
    register_transformer("t_x", MockTransformer)
    b = _to_transformer("t_x")
    assert isinstance(b, MockTransformer)

    register_transformer("t_t3", t3)
    register_transformer("t_t4", t4)
    register_transformer("t_t5", t5)
    assert isinstance(_to_transformer("t_t3"), Transformer)
    assert isinstance(_to_transformer("t_t4", "*,x:int"), Transformer)
    assert isinstance(_to_transformer("t_t5"), Transformer)

    # schema: *
    def register_temp(df: pd.DataFrame) -> pd.DataFrame:
        return df

    t = _to_transformer("register_temp")
    assert isinstance(t, Transformer)
    assert not isinstance(t, MockTransformer)
    # registered alias has the highest priority
    register_transformer("register_temp", MockTransformer)
    t = _to_transformer("register_temp")
    assert isinstance(t, MockTransformer)

    # can't overwrite
    raises(
        KeyError,
        lambda: register_transformer(
            "register_temp", MockTransformer, on_dup=ParamDict.THROW
        ),
    )


def test__to_transformer_determinism():
    a = _to_transformer(t1, None)
    b = _to_transformer(t1, None)
    c = _to_transformer("t1", None)
    assert a is not b
    assert to_uuid(a) == to_uuid(b)
    assert a is not c
    assert to_uuid(a) == to_uuid(c)

    a = _to_transformer(t4, "*,b:int")
    b = _to_transformer("t4", "*,b:int")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)

    a = _to_transformer(t4, "a:int,b:int")
    b = _to_transformer("t4", Schema("a:int,b:int"))
    assert a is not b
    assert to_uuid(a) == to_uuid(b)

    a = _to_transformer(MockTransformer)
    b = _to_transformer("MockTransformer")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)

    a = _to_transformer(t10)
    b = _to_transformer("t10")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)


def test_to_transformer_validation():
    @transformer(["*", None, "b:int"], input_has=" a , b ")
    def tv1(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
        for r in df:
            r["b"] = 1
            yield r

    # input_has: a , b
    # schema: *,b:int
    def tv2(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
        for r in df:
            r["b"] = 1
            yield r

    class MockTransformerV(Transformer):
        @property
        def validation_rules(self):
            return {"input_is": "a:int,b:int"}

        def get_output_schema(self, df):
            pass

        def transform(self, df):
            pass

    a = _to_transformer(tv1, None)
    assert {"input_has": ["a", "b"]} == a.validation_rules
    b = _to_transformer(tv2, None)
    assert {"input_has": ["a", "b"]} == b.validation_rules
    c = _to_transformer(MockTransformerV)
    assert {"input_is": "a:int,b:int"} == c.validation_rules


def test_inside_class():
    class Test(object):
        # schema: *
        # input_is: a:int , b :int
        def t1(self, df: pd.DataFrame) -> pd.DataFrame:
            return df

    test = Test()
    a = _to_transformer(test.t1)
    assert isinstance(a, Transformer)
    assert {"input_is": "a:int,b:int"} == a.validation_rules


def test_parse_transformer():
    @parse_transformer.candidate(lambda x: isinstance(x, str) and x.startswith("(("))
    def _parse(obj):
        return MockTransformer(obj)

    a = _to_transformer("((abc")
    b = _to_transformer("((bc")
    c = _to_transformer("((abc")

    assert isinstance(a, MockTransformer)
    assert isinstance(b, MockTransformer)
    assert isinstance(c, MockTransformer)
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(b)


@transformer(["*", None, "b:int"])
def t1(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = 1
        yield r


@transformer([Schema("b:int"), "*"])
def t2(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = 1
        yield r


@transformer(Schema("a:int, b:int"))
def t3(df: Iterable[List[Any]]) -> Iterable[List[Any]]:
    for r in df:
        r += [1]
        yield r


def t4(df: Iterable[List[Any]]) -> Iterable[List[Any]]:
    for r in df:
        r += [1]
        yield r


# schema: *,b:int
def t5(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = 1
        yield r


def t6(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = 1
        yield r


# schema: *
def t7(df: pd.DataFrame) -> Iterable[pd.DataFrame]:
    yield df


# schema: *
def t8(df: Iterable[pd.DataFrame]) -> pd.DataFrame:
    return pd.concat(list(df))


# schema: *
def t9(df: pd.DataFrame) -> Iterable[pd.DataFrame]:
    yield df


# schema: *
def t10(df: pd.DataFrame, c: callable) -> pd.DataFrame:
    yield df


class MockTransformer(Transformer):
    def __init__(self, x=""):
        self._x = x

    def get_output_schema(self, df):
        pass

    def transform(self, df):
        pass

    def __uuid__(self) -> str:
        return to_uuid(super().__uuid__(), self._x)



from fugue.rpc import RPCFunc, to_rpc_handler
from pytest import raises
from triad import to_uuid
from copy import copy, deepcopy


def test_rpc_func():
    def f1(a: str) -> str:
        return "1"

    d1 = RPCFunc(f1)
    d2 = to_rpc_handler(f1)
    assert to_uuid(d1) == to_uuid(d2)
    assert to_uuid(d1) == to_uuid(to_rpc_handler(d1))
    assert "1" == d1("x")
    with raises(ValueError):
        RPCFunc(1)


def test_determinism():
    def _f1(a: str) -> str:
        return "1"

    assert to_uuid(RPCFunc(_f1)) == to_uuid(to_rpc_handler(_f1))
    assert to_uuid(RPCFunc(lambda x: x)) == to_uuid(RPCFunc(lambda x: x + 1))


def test_no_copy():
    class T(object):
        def __init__(self):
            self.n = 0

        def call(self, n: int) -> int:
            self.n += n
            return self.n

    t = T()
    d1 = RPCFunc(t.call)
    assert 10 == d1(10)
    assert 10 == t.n

    d2 = to_rpc_handler(t.call)
    d2(10)

    d3 = to_rpc_handler(d1)
    d3(10)
    assert 30 == t.n

    d4 = copy(d3)
    d4(10)

    d5 = deepcopy(d4)
    d5(10)
    assert 50 == t.n



import os

from fugue._utils.io import save_df as pd_save_df
from fugue.dataframe.utils import _df_eq as df_eq
from pytest import mark, raises
from triad.collections.fs import FileSystem
from triad.exceptions import InvalidOperationError

from fugue_dask._io import load_df, save_df
from fugue_dask.dataframe import DaskDataFrame


def test_parquet_io(tmpdir):
    df1 = DaskDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    df2 = DaskDataFrame([[[1, 2]]], "a:[int]")
    # {a:int} will become {a:long} because pyarrow lib has issue
    df3 = DaskDataFrame([[dict(a=1)]], "a:{a:long}")
    for df in [df1, df2, df3]:
        path = os.path.join(tmpdir, "a.parquet")
        save_df(df, path)
        actual = load_df(path)
        df_eq(df, actual, throw=True)

    save_df(df1, path)
    actual = load_df(path, columns=["b", "a"])
    df_eq(actual, [[2, "1"]], "b:int,a:str")
    actual = load_df(path, columns="b:str,a:int")
    df_eq(actual, [["2", 1]], "b:str,a:int")
    # can't specify wrong columns
    raises(Exception, lambda: load_df(path, columns="bb:str,a:int"))

    # load directory
    fs = FileSystem()
    for name in ["folder.parquet", "folder"]:
        folder = os.path.join(tmpdir, name)
        fs.makedirs(folder)
        f0 = os.path.join(folder, "_SUCCESS")
        f1 = os.path.join(folder, "1.parquet")
        f2 = os.path.join(folder, "3.parquet")
        fs.touch(f0)
        pd_save_df(df1, f1)
        pd_save_df(df1, f2)

    actual = load_df(folder, "parquet")
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # load multiple paths
    actual = load_df([f1, f2], "parquet")
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # load folder
    actual = load_df(folder, "parquet")
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    actual = load_df(os.path.join(tmpdir, "folder.parquet"))
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # load pattern
    actual = load_df(os.path.join(tmpdir, "folder", "*.parquet"))
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # overwrite folder with single file
    save_df(actual, os.path.join(tmpdir, "folder.parquet"), mode="overwrite")
    actual = load_df(os.path.join(tmpdir, "folder.parquet"))
    df_eq(actual, [["1", 2, 3], ["1", 2, 3]], "a:str,b:int,c:long")

    # overwrite = False
    raises(FileExistsError, lambda: save_df(df1, f1, mode="error"))
    raises(
        FileExistsError,
        lambda: save_df(df1, os.path.join(tmpdir, "folder.parquet"), mode="error"),
    )

    # wrong mode
    raises(NotImplementedError, lambda: save_df(df1, f1, mode="dummy"))


def test_csv_io(tmpdir):
    df1 = DaskDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    path = os.path.join(tmpdir, "a.csv")
    # without header
    save_df(df1, path)
    # assert fs.readtext(path).startswith("1,2,3")
    raises(ValueError, lambda: load_df(path, header=False))
    actual = load_df(path, columns=["a", "b", "c"], header=False, infer_schema=True)
    assert [[1, 2, 3]] == actual.as_array()
    assert actual.schema == "a:long,b:long,c:long"
    actual = load_df(path, columns="a:double,b:str,c:str", header=False)
    assert [[1.0, "2", "3"]] == actual.as_array()
    assert actual.schema == "a:double,b:str,c:str"
    # with header
    save_df(df1, path, header=True)
    # assert fs.readtext(path).startswith("a,b,c")
    actual = load_df(path, header=True)
    assert [["1", "2", "3"]] == actual.as_array()
    actual = load_df(path, header=True, infer_schema=True)
    assert [[1, 2, 3]] == actual.as_array()
    actual = load_df(path, columns=["b", "a"], header=True, infer_schema=True)
    assert [[2, 1]] == actual.as_array()
    actual = load_df(path, columns="b:str,a:double", header=True)
    assert [["2", 1.0]] == actual.as_array()
    raises(KeyError, lambda: load_df(path, columns="b:str,x:double", header=True))

    raises(
        NotImplementedError, lambda: load_df(path, columns="b:str,x:double", header=2)
    )


def test_json(tmpdir):
    df1 = DaskDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    path = os.path.join(tmpdir, "a.json")
    save_df(df1, path)
    actual = load_df(path)
    df_eq(actual, [[1, 2, 3]], "a:long,b:long,c:long")
    actual = load_df(path, columns=["b", "a"])
    df_eq(actual, [[2, "1"]], "b:int,a:str")
    actual = load_df(path, columns="b:str,a:int")
    df_eq(actual, [["2", 1]], "b:str,a:int")
    raises(KeyError, lambda: load_df(path, columns="bb:str,a:int"))


@mark.skip(reason="Unable to test due to spark jars not being downloaded properly")
def test_avro_io(tmpdir):
    df1 = DaskDataFrame([["1", 2, 3]], "a:str,b:int,c:long")
    path = os.path.join(tmpdir, "a.avro")
    save_df(df1, path)
    actual = load_df(path)

    df_eq(actual, [["1", 2, 3]], "a:str,b:long,c:long")
    actual = load_df(path, columns=["a", "b"])
    df_eq(actual, [["1", 3]], "a:str,b:long")

    actual = load_df(path, columns="a:str,b:int,c:long")
    df_eq(actual, [["1", 2, 3]], "a:str,b:int,c:long")

    actual = load_df(path, columns=["b", "c"], infer_schema=True)
    df_eq(actual, [[2, 3]], "b:long,c:long")

    # provide schema and columns -> throw error
    raises(
        Exception,
        lambda: save_df(
            path,
            columns="a:str,b:int,c:long",
            schema={
                "type": "record",
                "name": "Root",
                "fields": [
                    {"name": "station", "type": "string"},
                    {"name": "time", "type": "long"},
                    {"name": "temp", "type": "int"},
                ],
            },
        ),
    )

    # provide schema and infer_schema is True -> throw error
    raises(
        Exception,
        lambda: save_df(
            path,
            columns=None,
            schema={
                "type": "record",
                "name": "Root",
                "fields": [
                    {"name": "station", "type": "string"},
                    {"name": "time", "type": "long"},
                    {"name": "temp", "type": "int"},
                ],
            },
            infer_schema=True,
        ),
    )



from typing import Any, Dict, Iterable, List, Optional

import pandas as pd
import pyarrow as pa
from duckdb import DuckDBPyRelation
from triad import Schema

from fugue import ArrayDataFrame, ArrowDataFrame, DataFrame, LocalBoundedDataFrame
from fugue.exceptions import FugueDataFrameOperationError, FugueDatasetEmptyError
from fugue.plugins import (
    as_fugue_dataset,
    as_local_bounded,
    get_column_names,
    get_num_partitions,
    is_df,
)

from ._utils import encode_column_name, to_duck_type, to_pa_type


class DuckDataFrame(LocalBoundedDataFrame):
    """DataFrame that wraps DuckDB ``DuckDBPyRelation``.

    :param rel: ``DuckDBPyRelation`` object
    """

    def __init__(self, rel: DuckDBPyRelation):
        self._rel = rel
        super().__init__(schema=self._get_schema)

    def _get_schema(self) -> Schema:
        return Schema(
            [
                pa.field(x, to_pa_type(y))
                for x, y in zip(self._rel.columns, self._rel.types)
            ]
        )

    @property
    def alias(self) -> str:
        return "_" + str(id(self._rel))  # DuckDBPyRelation.alias is not always unique

    @property
    def native(self) -> DuckDBPyRelation:
        """DuckDB relation object"""
        return self._rel

    def native_as_df(self) -> DuckDBPyRelation:
        return self._rel

    @property
    def empty(self) -> bool:
        return self.count() == 0

    def peek_array(self) -> List[Any]:
        res = self._rel.limit(1).to_df()
        if res.shape[0] == 0:
            raise FugueDatasetEmptyError()
        return res.values.tolist()[0]

    def count(self) -> int:
        return len(self._rel)

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        cols = [col for col in self._rel.columns if col not in cols]
        rel = self._rel.project(",".join(encode_column_name(n) for n in cols))
        return DuckDataFrame(rel)

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        rel = self._rel.project(",".join(encode_column_name(n) for n in keys))
        return DuckDataFrame(rel)

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        _assert_no_missing(self._rel, columns.keys())
        expr = ", ".join(
            f"{a} AS {b}"
            for a, b in [
                (encode_column_name(name), encode_column_name(columns.get(name, name)))
                for name in self._rel.columns
            ]
        )
        return DuckDataFrame(self._rel.project(expr))

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self._get_altered_schema(columns)
        if new_schema == self.schema:
            return self
        fields: List[str] = []
        for f1, f2 in zip(self.schema.fields, new_schema.fields):
            if f1.type == f2.type:
                fields.append(encode_column_name(f1.name))
            else:
                tp = to_duck_type(f2.type)
                fields.append(
                    f"CAST({encode_column_name(f1.name)} AS {tp}) "
                    f"AS {encode_column_name(f1.name)}"
                )
        return DuckDataFrame(self._rel.project(", ".join(fields)))

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        return self._rel.arrow()

    def as_pandas(self) -> pd.DataFrame:
        if any(pa.types.is_nested(f.type) for f in self.schema.fields):
            # Duckdb has issue to directly convert nested types to pandas
            return ArrowDataFrame(self.as_arrow()).as_pandas()
        return self._rel.to_df()

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        res = ArrowDataFrame(self.as_arrow())
        if self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        if columns is not None:
            return self[columns].as_array(type_safe=type_safe)
        return self._fetchall(self._rel)

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if columns is not None:
            yield from self[columns].as_array_iterable(type_safe=type_safe)
        else:
            yield from self._fetchall(self._rel)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        if columns is not None:
            return self[columns].head(n)
        return ArrayDataFrame(self._fetchall(self._rel.limit(n)), schema=self.schema)

    def _fetchall(self, rel: DuckDBPyRelation) -> List[List[Any]]:
        map_pos = [i for i, t in enumerate(self.schema.types) if pa.types.is_map(t)]
        if len(map_pos) == 0:
            return [list(x) for x in rel.fetchall()]
        else:

            def to_list(row: Any) -> List[Any]:
                res = list(row)
                for p in map_pos:
                    res[p] = list(zip(row[p]["key"], row[p]["value"]))
                return res

            return [to_list(x) for x in rel.fetchall()]


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, DuckDBPyRelation))
def _duckdb_as_fugue_df(df: DuckDBPyRelation, **kwargs: Any) -> DuckDataFrame:
    return DuckDataFrame(df, **kwargs)


@is_df.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duck_is_df(df: DuckDBPyRelation) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duckdb_num_partitions(df: DuckDBPyRelation) -> int:
    return 1


@as_local_bounded.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _duck_as_local(df: DuckDBPyRelation) -> DuckDBPyRelation:
    return df


@get_column_names.candidate(lambda df: isinstance(df, DuckDBPyRelation))
def _get_duckdb_columns(df: DuckDBPyRelation) -> List[Any]:
    return list(df.columns)


def _assert_no_missing(df: DuckDBPyRelation, columns: Iterable[Any]) -> None:
    missing = set(columns) - set(df.columns)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")



from fugue import FugueWorkflow
from fugue import fsql


def test_importless():
    for engine in ["duck", "duckdb"]:
        dag = FugueWorkflow()
        dag.df([[0]], "a:int").show()

        dag.run(engine)

        fsql(
            """
        CREATE [[0],[1]] SCHEMA a:int
        SELECT * WHERE a<1
        PRINT
        """
        ).run(engine)

        dag = FugueWorkflow()
        idf = dag.df([[0], [1]], "a:int").as_ibis()
        idf[idf.a < 1].as_fugue().show()

        dag.run(engine)

        dag = FugueWorkflow()
        tdf = dag.df([[0], [1]], "a:int")
        dag.select("SELECT * FROM ", tdf, " WHERE a<1", sql_engine=engine)

        dag.run()



"""
All modeuls for developing and extending Fugue
"""
# flake8: noqa
# pylint: disable-all

from triad.collections.function_wrapper import AnnotatedParam

from fugue.bag.bag import BagDisplay
from fugue.collections.partition import PartitionCursor, PartitionSpec
from fugue.collections.sql import StructuredRawSQL, TempTableName
from fugue.collections.yielded import PhysicalYielded, Yielded
from fugue.dataframe.function_wrapper import (
    DataFrameFunctionWrapper,
    DataFrameParam,
    LocalDataFrameParam,
    fugue_annotated_param,
)
from fugue.dataset import DatasetDisplay
from fugue.execution.execution_engine import (
    EngineFacet,
    ExecutionEngineParam,
    MapEngine,
    SQLEngine,
)
from fugue.execution.factory import (
    is_pandas_or,
    make_execution_engine,
    make_sql_engine,
    register_default_execution_engine,
    register_default_sql_engine,
    register_execution_engine,
    register_sql_engine,
)
from fugue.execution.native_execution_engine import PandasMapEngine, QPDPandasEngine
from fugue.rpc import (
    EmptyRPCHandler,
    RPCClient,
    RPCFunc,
    RPCHandler,
    RPCServer,
    make_rpc_server,
    to_rpc_handler,
)
from fugue.workflow._workflow_context import FugueWorkflowContext
from fugue.workflow.module import module
from fugue.workflow.workflow import FugueWorkflow, WorkflowDataFrame, WorkflowDataFrames



from typing import Any, Dict, Iterable, List

from fugue.dataframe import ArrayDataFrame
from fugue.exceptions import FugueInterfacelessError
from fugue.extensions.transformer import Transformer, _to_transformer, transformer
from pytest import raises
from triad.collections.schema import Schema


def test_transformer():
    assert isinstance(t1, Transformer)
    df = ArrayDataFrame([[0]], "a:int")
    t1._output_schema = t1.get_output_schema(df)
    assert t1.output_schema == "a:int,b:int"
    t2._output_schema = t2.get_output_schema(df)
    assert t2.output_schema == "b:int,a:int"
    assert [[0, 1]] == list(t3(df.as_array_iterable()))


def test__to_transformer():
    a = _to_transformer(t1, None)
    assert isinstance(a, Transformer)
    a._x = 1
    # every parse should produce a different transformer even the input is
    # a transformer instance
    b = _to_transformer(t1, None)
    assert isinstance(b, Transformer)
    assert "_x" not in b.__dict__
    c = _to_transformer("t1", None)
    assert isinstance(c, Transformer)
    assert "_x" not in c.__dict__
    c._x = 1
    d = _to_transformer("t1", None)
    assert isinstance(d, Transformer)
    assert "_x" not in d.__dict__
    raises(FugueInterfacelessError, lambda: _to_transformer(t4, None))
    raises(FugueInterfacelessError, lambda: _to_transformer("t4", None))
    e = _to_transformer("t4", "*,b:int")
    assert isinstance(e, Transformer)


@transformer(["*", None, "b:int"])
def t1(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = 1
        yield r


@transformer([Schema("b:int"), "*"])
def t2(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = 1
        yield r


@transformer("*, b:int")
def t3(df: Iterable[List[Any]]) -> Iterable[List[Any]]:
    for r in df:
        r += [1]
        yield r


def t4(df: Iterable[List[Any]]) -> Iterable[List[Any]]:
    for r in df:
        r += [1]
        yield r



import ray
from triad import Schema

from fugue_ray import RayDataFrame
from fugue_ray._utils.dataframe import add_partition_key


def test_add_partition_key():
    with ray.init():
        df = RayDataFrame([[0, "a"], [1, "b"]], "a:int,b:str")
        res, s = add_partition_key(df.native, df.schema, ["b", "a"], output_key="x")
        assert s == Schema("a:int,b:str,x:binary")

        res, s = add_partition_key(df.native, df.schema, ["b"], output_key="x")
        assert s == "a:int,b:str,x:str"
        assert RayDataFrame(res, s).as_array() == [[0, "a", "a"], [1, "b", "b"]]



import json

from fugue_sql_antlr import FugueSQLParser
from pytest import raises
from triad.collections.schema import Schema

from fugue.collections.partition import PartitionSpec
from fugue.dataframe import DataFrame, DataFrames, LocalDataFrame
from fugue.sql._visitors import _Extensions, _VisitorBase
from fugue.exceptions import FugueSQLSyntaxError
from fugue.workflow.workflow import FugueWorkflow

_PARSE_MODE = "auto"


def test_json():
    def assert_eq(expr, expected):
        sql = FugueSQLParser(
            expr,
            "fugueJsonValue",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = v.visit(sql.tree)
        if expected is None:
            assert obj is None
        else:
            assert expected == obj

    # int
    assert_eq("1", 1)
    # str
    assert_eq("'1'", "1")
    assert_eq('"1"', "1")
    assert_eq('"\\"x"', '"x')
    assert_eq("'\\'x'", "'x")
    # bool
    assert_eq("True", True)
    assert_eq("falsE", False)
    assert_eq("falsE", False)
    # null
    assert_eq("Null", None)
    # dict
    assert_eq("{}", dict())
    assert_eq("{a:1\n}", {"a": 1})
    assert_eq("{aA:1.5,}", {"aA": 1.5})
    assert_eq("{'a':.5}", {"a": 0.5})
    assert_eq("('a':.5)", {"a": 0.5})
    assert_eq("(a:1,b=2)", {"a": 1, "b": 2})
    # array
    assert_eq("[]", list())
    assert_eq("[\n1,'a'\n]", [1, "a"])
    assert_eq("[\n1,'a'\n,]", [1, "a"])
    # python arg like
    assert_eq("(a=1,b='x',c=True,d={x:3},)", dict(a=1, b="x", c=True, d=dict(x=3)))


def test_extension():
    def assert_eq(expr, expected=None):
        sql = FugueSQLParser(
            expr,
            "fugueExtension",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = v.visit(sql.tree)
        assert expected == obj

    assert_eq("abc", "abc")
    assert_eq("abc . def", "abc.def")
    assert_eq("abc . def. ` ` ", "abc.def.` `")
    assert_eq(" x :abc . def. ` ` ", ("x", "abc.def.` `"))
    assert_eq("`x:y ` : abc . def. ` ` ", ("x:y ", "abc.def.` `"))


def test_schema():
    def assert_eq(expr, expected=None):
        sql = FugueSQLParser(
            expr,
            "fugueSchema",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = v.visit(sql.tree)
        if expected is None:
            expected = expr
        assert Schema(expected) == obj

    assert_eq("a:int,b:str")
    assert_eq("a:int,b:string", "a:int,b:str")
    assert_eq("a:int,B:str,c:[int]")
    assert_eq("a:int,b:str,C:{A:int,b:str}")
    assert_eq("a:int,\nb:str,C:[{A:int,b:str}]")
    assert_eq("a:<int,str>,b:<int,[str]>,c:[<int,{b:str}>]")
    assert_eq(" ` * ` : int , `` : string ", "` * `:int,``:str")


def test_wild_schema():
    def assert_eq(expr, expected=None):
        sql = FugueSQLParser(
            expr,
            "fugueWildSchema",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = v.visit(sql.tree)
        if expected is None:
            expected = expr
        assert expected == obj

    assert_eq("a:int,b:str")
    assert_eq("a:int, b: string", "a:int,b:str")
    assert_eq("a:int32, *", "a:int,*")
    assert_eq("*  ", "*")
    assert_eq("*,a:int32", "*,a:int")
    assert_eq("a:int32, *, \nb:string", "a:int,*,b:str")
    # transformations
    assert_eq(
        "a:int32, *, B:string +x:str, k:int\n-Y, z -t~ w , x",
        "a:int,*,B:str+x:str,k:int-Y,z-t~w,x",
    )
    # special chars
    assert_eq(
        " ` * ` :int32, *, `` : string +x:str, k:int\n-Y, z -t~ w , x",
        "` * `:int,*,``:str+x:str,k:int-Y,z-t~w,x",
    )


def test_pre_partition():
    def assert_eq(expr, expected):
        sql = FugueSQLParser(
            expr,
            "fuguePrepartition",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = json.dumps(v.visit(sql.tree).jsondict)
        assert json.dumps(expected.jsondict) == obj

    assert_eq("prepartition 100", PartitionSpec(num=100))
    assert_eq(
        "prepartition ROWCOUNT*\n3+ (12.5-CONCUrrency/2)",
        PartitionSpec(num="ROWCOUNT*3+(12.5-CONCURRENCY/2)"),
    )
    assert_eq("prepartition by a, b", PartitionSpec(by=["a", "b"]))
    assert_eq(
        "HASH PrePARTITION 100 BY a,b",
        PartitionSpec(algo="hash", num=100, by=["a", "b"]),
    )
    assert_eq(
        "EVEN prepartition 100 BY a,b",
        PartitionSpec(algo="even", num=100, by=["a", "b"]),
    )
    assert_eq(
        "rand prepartition 100 BY a,b",
        PartitionSpec(algo="rand", num=100, by=["a", "b"]),
    )
    assert_eq(
        "prepartition 100 presort a,\nb\ndesc",
        PartitionSpec(num=100, presort="a asc, b desc"),
    )


def test_params():
    def assert_eq(expr, expected):
        sql = FugueSQLParser(
            expr,
            "fugueParams",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = v.visit(sql.tree)
        assert expected == obj

    assert_eq("params a:10,b:'x'", dict(a=10, b="x"))
    assert_eq("params {a:10,b:'x'}", dict(a=10, b="x"))
    assert_eq("params (a:10,b:'x')", dict(a=10, b="x"))
    assert_eq("{a:10,b:'x'}", dict(a=10, b="x"))
    assert_eq("(a=1,b='x',c=True,d={x:3},)", dict(a=1, b="x", c=True, d=dict(x=3)))


def test_single_output_common_expr():
    def assert_eq(expr, using, params, schema):
        sql = FugueSQLParser(
            expr,
            "fugueSingleOutputExtensionCommon",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = v.visit(sql.tree)
        assert using == obj["fugueUsing"]
        if params is None:
            assert "params" not in obj
        else:
            assert params == obj["params"]
        if schema is None:
            assert "schema" not in obj
        else:
            assert obj["schema"] == schema

    assert_eq("using a.B.c", "a.B.c", None, None)
    assert_eq("using a.B.c()", "a.B.c", {}, None)
    assert_eq("using a.B.c{}", "a.B.c", {}, None)
    assert_eq("using a.B.c(a=1,b=2)", "a.B.c", dict(a=1, b=2), None)
    assert_eq(
        "using a.B.c params a:1,b:2 schema a:int,b:int",
        "a.B.c",
        dict(a=1, b=2),
        "a:int,b:int",
    )


def test_assignment():
    def assert_eq(expr, varname, sign):
        sql = FugueSQLParser(
            expr,
            "fugueAssignment",
            ignore_case=True,
            parse_mode=_PARSE_MODE,
        )
        v = _VisitorBase(sql)
        obj = v.visit(sql.tree)
        assert (varname, sign) == obj

    assert_eq("aA=", "aA", "=")
    # assert_eq("aA??", "aA", "??")



from typing import Any, Callable, Dict, Iterable, List

import pandas as pd
from fugue.extensions.transformer import (
    Transformer,
    _to_output_transformer,
    output_transformer,
    parse_output_transformer,
    register_output_transformer,
)
from fugue.extensions.transformer.constants import OUTPUT_TRANSFORMER_DUMMY_SCHEMA
from triad.utils.hash import to_uuid


def test_transformer():
    assert isinstance(t1, Transformer)
    assert t1.get_output_schema(None) == OUTPUT_TRANSFORMER_DUMMY_SCHEMA


def test__to_output_transformer():
    a = _to_output_transformer(MockTransformer)
    assert isinstance(a, MockTransformer)
    b = _to_output_transformer("MockTransformer")
    assert isinstance(b, MockTransformer)

    a = _to_output_transformer(t1)
    assert isinstance(a, Transformer)
    a._x = 1
    # every parse should produce a different transformer even the input is
    # a transformer instance
    b = _to_output_transformer(t1)
    assert isinstance(b, Transformer)
    assert "_x" not in b.__dict__
    c = _to_output_transformer("t1")
    assert isinstance(c, Transformer)
    assert "_x" not in c.__dict__
    c._x = 1
    d = _to_output_transformer("t1")
    assert isinstance(d, Transformer)
    assert "_x" not in d.__dict__
    e = _to_output_transformer("t4")
    assert isinstance(e, Transformer)
    f = _to_output_transformer("t5")
    assert isinstance(f, Transformer)
    g = _to_output_transformer("t6")
    assert isinstance(g, Transformer)
    h = _to_output_transformer("t7")
    assert isinstance(h, Transformer)
    i = _to_output_transformer("t8")
    assert isinstance(i, Transformer)


def test__register():
    register_output_transformer("ot_x", MockTransformer)
    b = _to_output_transformer("ot_x")
    assert isinstance(b, MockTransformer)


def test__to_output_transformer_determinism():
    a = _to_output_transformer(t1)
    b = _to_output_transformer(t1)
    c = _to_output_transformer("t1")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)
    assert a is not c
    assert to_uuid(a) == to_uuid(c)

    a = _to_output_transformer(t4)
    b = _to_output_transformer("t4")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)

    a = _to_output_transformer(MockTransformer)
    b = _to_output_transformer("MockTransformer")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)

    a = _to_output_transformer(t8)
    b = _to_output_transformer("t8")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)


def test_to_output_transformer_validation():
    @output_transformer(input_has=" a , b ")
    def tv1(df: Iterable[Dict[str, Any]]) -> None:
        pass

    # input_has: a , b
    def tv2(df: Iterable[Dict[str, Any]]) -> None:
        pass

    class MockTransformerV(Transformer):
        @property
        def validation_rules(self):
            return {"input_is": "a:int,b:int"}

        def get_output_schema(self, df):
            pass

        def transform(self, df):
            pass

    a = _to_output_transformer(tv1, None)
    assert {"input_has": ["a", "b"]} == a.validation_rules
    b = _to_output_transformer(tv2, None)
    assert {"input_has": ["a", "b"]} == b.validation_rules
    c = _to_output_transformer(MockTransformerV)
    assert {"input_is": "a:int,b:int"} == c.validation_rules


def test_parse_output_transformer():
    @parse_output_transformer.candidate(
        lambda x: isinstance(x, str) and x.startswith("((")
    )
    def _parse(obj):
        return MockTransformer(obj)

    a = _to_output_transformer("((abc")
    b = _to_output_transformer("((bc")
    c = _to_output_transformer("((abc")

    assert isinstance(a, MockTransformer)
    assert isinstance(b, MockTransformer)
    assert isinstance(c, MockTransformer)
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(b)


@output_transformer()
def t1(df: Iterable[Dict[str, Any]]) -> None:
    pass


def t4(df: Iterable[List[Any]]) -> None:
    pass


# schema: *,b:int
def t5(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for r in df:
        r["b"] = 1
        yield r


def t6(df: Iterable[List[Any]], **kwargs) -> None:
    pass


# Consistency with transformer
def t7(df: pd.DataFrame) -> Iterable[pd.DataFrame]:
    pass


def t8(df: pd.DataFrame, c: Callable[[str], str]) -> Iterable[pd.DataFrame]:
    pass


class MockTransformer(Transformer):
    def __init__(self, x=""):
        self._x = x

    def get_output_schema(self, df):
        pass

    def transform(self, df):
        pass

    def __uuid__(self) -> str:
        return to_uuid(super().__uuid__(), self._x)



# flake8: noqa
from triad import run_at_def

from ._compat import IbisSchema, IbisTable
from .dataframe import IbisDataFrame
from .execution.ibis_engine import IbisEngine, parse_ibis_engine
from .execution.pandas_backend import PandasIbisEngine
from .execution_engine import IbisExecutionEngine, IbisSQLEngine
from .extensions import as_fugue, as_ibis, run_ibis



import sys

import pytest
from fugue_test.builtin_suite import BuiltInTests
from fugue_test.execution_suite import ExecutionEngineTests

from .mock.execution_engine import MockDuckExecutionEngine


@pytest.mark.skipif(sys.version_info < (3, 8), reason="< 3.8")
class IbisExecutionEngineTests(ExecutionEngineTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._engine = cls.make_engine(cls)

    @classmethod
    def tearDownClass(cls):
        # cls._con.close()
        pass

    def make_engine(self):
        return MockDuckExecutionEngine({"test": True})

    def test_select(self):
        # it can't work properly with DuckDB (hugeint is not recognized)
        pass


@pytest.mark.skipif(sys.version_info < (3, 8), reason="< 3.8")
class IbisExecutionEngineForceIbisTests(ExecutionEngineTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._engine = cls.make_engine(cls)

    @classmethod
    def tearDownClass(cls):
        # cls._con.close()
        pass

    def make_engine(self):
        return MockDuckExecutionEngine({"test": True}, force_is_ibis=True)

    def test_properties(self):
        assert not self.engine.is_distributed
        assert not self.engine.map_engine.is_distributed
        assert not self.engine.sql_engine.is_distributed

    def test_select(self):
        # it can't work properly with DuckDB (hugeint is not recognized)
        pass

    def test_get_parallelism(self):
        assert self.engine.get_current_parallelism() == 1


@pytest.mark.skipif(sys.version_info < (3, 8), reason="< 3.8")
class DuckBuiltInTests(BuiltInTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._engine = cls.make_engine(cls)

    @classmethod
    def tearDownClass(cls):
        # cls._con.close()
        pass

    def make_engine(self):
        return MockDuckExecutionEngine({"test": True})

    def test_df_select(self):
        # it can't work properly with DuckDB (hugeint is not recognized)
        pass


@pytest.mark.skipif(sys.version_info < (3, 8), reason="< 3.8")
class DuckBuiltInForceIbisTests(BuiltInTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._engine = cls.make_engine(cls)

    @classmethod
    def tearDownClass(cls):
        # cls._con.close()
        pass

    def make_engine(self):
        return MockDuckExecutionEngine({"test": True}, force_is_ibis=True)

    def test_df_select(self):
        # it can't work properly with DuckDB (hugeint is not recognized)
        pass



from fugue.sql._utils import fill_sql_template


def test_fill_sql_template():
    data = {"a": 1, "b": "x"}
    assert "select * from tbl where a = 1 and b = 'x'" == fill_sql_template(
        "select * from tbl where a = {{a}} and b = '{{b}}'", data
    )
    assert """select * from tbl where a = 1 and b = "x" """ == fill_sql_template(
        """select * from tbl where a = {{a}} and b = "{{b}}" """, data
    )

    assert """select * where b="%x" """ == fill_sql_template(
        """select * where b="%{{b}}" """, data
    )
    assert """select * where b="x%" """ == fill_sql_template(
        """select * where b="{{b}}%" """, data
    )

    assert """select * b like "{}%{}" """ == fill_sql_template(
        """select * b like "{}%{}" """, data
    )
    assert """select * b like '%}' """ == fill_sql_template(
        """select * b like '%}' """, data
    )
    assert """select * where b="%x" """ == fill_sql_template(
        """select * where b="%{{b}}" """, data
    )
    assert """select * where b="x%" """ == fill_sql_template(
        """select * where b="{{b}}%" """, data
    )

    assert "a=select " == fill_sql_template("a=select ", data)

    # try single quotes for finding json patterns
    assert "a=select * from b like '{%'" == fill_sql_template(
        "a=select * from b like '{%'", data
    )
    assert "a=select * from b like '%}'" == fill_sql_template(
        "a=select * from b like '%}'", data
    )

    # try double quotes for finding json patterns
    assert 'a=select * from b like "%}"' == fill_sql_template(
        'a=select * from b like "%}"', data
    )
    assert 'a=select * from b like "{%"' == fill_sql_template(
        'a=select * from b like "{%"', data
    )

    assert "1x1" == fill_sql_template("{{a}}{{b}}{{a}}", data)
    assert "" == fill_sql_template("", data)
    assert "%s" == fill_sql_template("%s", data)
    assert "%%s" == fill_sql_template("%%s", data)
    assert "1%%sx1" == fill_sql_template("{{a}}%%s{{b}}{{a}}", data)
    assert "1" == fill_sql_template("{{a}}", {"a": 1, "self": 2})


def test_fill_sql_template_array():
    data = {"a": [0, 1, 2]}
    assert """select * from tbl where a in ('0','1','2')""" == fill_sql_template(
        """select * from tbl where a in (
            {%- for i in a -%}
                {%- if loop.index0 < loop.length - 1 -%}'{{-i-}}',
                {%- else -%}'{{-i-}}'
                {%- endif -%}
            {%- endfor -%}
            )""",
        data,
    )

    def upper(word):
        return word.upper()

    data = {"a": ["a", "b", "c"]}
    assert """select * from tbl where a in ('A','B','C')""" == fill_sql_template(
        """select * from tbl where a in (
            {%- for i in a -%}
                {%- if loop.index0 < loop.length - 1 -%}'{{-i|upper-}}',
                {%- else -%}'{{-i|upper-}}'
                {%- endif -%}
            {%- endfor -%}
            )""",
        data,
    )
